# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from unittest import TestCase

from num2words import num2words


class Num2WordsSRTest(TestCase):

    def test_cardinal(self):
        self.assertEqual("sto", num2words(100, lang='sr'))
        self.assertEqual("sto jedan", num2words(101, lang='sr'))
        self.assertEqual("sto deset", num2words(110, lang='sr'))
        self.assertEqual("sto petnaest", num2words(115, lang='sr'))
        self.assertEqual(
            "sto dvadeset tri", num2words(123, lang='sr')
        )
        self.assertEqual(
            "jedna hiljada", num2words(1000, lang='sr')
        )
        self.assertEqual(
            "jedna hiljada jedan", num2words(1001, lang='sr')
        )
        self.assertEqual(
            "dve hiljade dvanaest", num2words(2012, lang='sr')
        )
        self.assertEqual(
            "dvanaest hiljada petsto devetnaest zapeta osamdeset pet",
            num2words(12519.85, lang='sr')
        )
        self.assertEqual(
            "jedan bilion dvesta trideset četiri miliona petsto "
            "šezdeset sedam hiljada osamsto devedeset",
            num2words(1234567890, lang='sr')
        )
        self.assertEqual(
            "dvesta petnaest noniliona četristo šezdeset jedan "
            "oktilion četristo sedam septiliona osamsto devedeset "
            "dva sekstiliona trideset devet kvintiliona dva kvadriliona "
            "sto pedeset sedam triliona sto osamdeset devet biliona "
            "osamsto osamdeset tri miliona devetsto jedna hiljada "
            "šesto sedamdeset šest",
            num2words(215461407892039002157189883901676, lang='sr')
        )
        self.assertEqual(
            "sedamsto devetnaest noniliona devedeset četiri oktiliona "
            "dvesta trideset četiri septiliona šesto devedeset tri "
            "sekstiliona šesto šezdeset tri kvintiliona trideset "
            "četiri kvadriliona osamsto dvadeset dva triliona osamsto "
            "dvadeset četiri biliona trista osamdeset četiri miliona "
            "dvesta dvadeset hiljada dvesta devedeset jedan",
            num2words(719094234693663034822824384220291, lang='sr')
        )
        self.assertEqual("pet", num2words(5, lang='sr'))
        self.assertEqual("petnaest", num2words(15, lang='sr'))
        self.assertEqual("sto pedeset četiri", num2words(154, lang='sr'))
        self.assertEqual(
            "jedna hiljada sto trideset pet",
            num2words(1135, lang='sr')
        )
        self.assertEqual(
            "četristo osamnaest hiljada petsto trideset jedan",
            num2words(418531, lang='sr'),
        )
        self.assertEqual(
            "jedan milion sto trideset devet",
            num2words(1000139, lang='sr')
        )

    def test_floating_point(self):
        self.assertEqual("pet zapeta dva", num2words(5.2, lang='sr'))
        self.assertEqual(
            num2words(10.02, lang='sr'),
            "deset zapeta nula dva"
        )
        self.assertEqual(
            num2words(15.007, lang='sr'),
            "petnaest zapeta nula nula sedam"
        )
        self.assertEqual(
            "petsto šezdeset jedan zapeta četrdeset dva",
            num2words(561.42, lang='sr')
        )

    def test_to_ordinal(self):
        # @TODO: implement to_ordinal
        with self.assertRaises(NotImplementedError):
            num2words(1, lang='sr', to='ordinal')

    def test_to_currency(self):
        self.assertEqual(
            'jedan evro, nula centi',
            num2words(1.0, lang='sr', to='currency', currency='EUR')

        )
        self.assertEqual(
            'dva evra, nula centi',
            num2words(2.0, lang='sr', to='currency', currency='EUR')

        )
        self.assertEqual(
            'pet evra, nula centi',
            num2words(5.0, lang='sr', to='currency', currency='EUR')

        )
        self.assertEqual(
            'dva evra, jedan cent',
            num2words(2.01, lang='sr', to='currency', currency='EUR')

        )

        self.assertEqual(
            'dva evra, dva centa',
            num2words(2.02, lang='sr', to='currency', currency='EUR')

        )
        self.assertEqual(
            'dva evra, pet centi',
            num2words(2.05, lang='sr', to='currency', currency='EUR')

        )
        self.assertEqual(
            'dve rublje, nula kopejki',
            num2words(2.0, lang='sr', to='currency', currency='RUB')

        )
        self.assertEqual(
            'dve rublje, jedna kopejka',
            num2words(2.01, lang='sr', to='currency', currency='RUB')

        )
        self.assertEqual(
            'dve rublje, dve kopejke',
            num2words(2.02, lang='sr', to='currency', currency='RUB')

        )
        self.assertEqual(
            'dve rublje, pet kopejki',
            num2words(2.05, lang='sr', to='currency', currency='RUB')

        )
        self.assertEqual(
            'jedan dinar, nula para',
            num2words(1.0, lang='sr', to='currency', currency='RSD')
        )
        self.assertEqual(
            'dva dinara, dve pare',
            num2words(2.02, lang='sr', to='currency', currency='RSD')

        )
        self.assertEqual(
            'pet dinara, pet para',
            num2words(5.05, lang='sr', to='currency', currency='RSD')

        )
        self.assertEqual(
            'jedanaest dinara, jedanaest para',
            num2words(11.11, lang='sr', to='currency', currency='RSD')

        )
        self.assertEqual(
            'dvadeset jedan dinar, dvadeset jedna para',
            num2words(21.21, lang='sr', to='currency', currency='RSD')

        )
        self.assertEqual(
            'dvadeset jedan evro, dvadeset jedan cent',
            num2words(21.21, lang='sr', to='currency', currency='EUR')

        )
        self.assertEqual(
            'dvadeset jedna rublja, dvadeset jedna kopejka',
            num2words(21.21, lang='sr', to='currency', currency='RUB')

        )
        self.assertEqual(
            'jedna hiljada dvesta trideset četiri evra, '
            'pedeset šest centi',
            num2words(
                1234.56, lang='sr', to='currency', currency='EUR'
            )
        )
        self.assertEqual(
            'jedna hiljada dvesta trideset četiri rublje, '
            'pedeset šest kopejki',
            num2words(
                1234.56, lang='sr', to='currency', currency='RUB'
            )
        )
        self.assertEqual(
            'sto jedan evro i jedanaest centi',
            num2words(
                10111,
                lang='sr',
                to='currency',
                currency='EUR',
                separator=' i'
            )
        )
        self.assertEqual(
            'sto jedna rublja i dvadeset jedna kopejka',
            num2words(
                10121,
                lang='sr',
                to='currency',
                currency='RUB',
                separator=' i'
            )
        )
        self.assertEqual(
            'sto jedna rublja i dvadeset dve kopejke',
            num2words(10122, lang='sr', to='currency', currency='RUB',
                      separator=' i')
        )
        self.assertEqual(
            'sto jedan evro i dvadeset jedan cent',
            num2words(10121, lang='sr', to='currency', currency='EUR',
                      separator=' i'),
        )
        self.assertEqual(
            'minus dvanaest hiljada petsto devetnaest evra, 85 centi',
            num2words(
                -1251985,
                lang='sr',
                to='currency',
                currency='EUR',
                cents=False
            )
        )
        self.assertEqual(
            "trideset osam evra i 40 centi",
            num2words('38.4', lang='sr', to='currency', separator=' i',
                      cents=False, currency='EUR'),
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

import math
from collections import OrderedDict
from decimal import Decimal

from .compat import to_s
from .currency import parse_currency_parts, prefix_currency


class Num2Word_Base(object):
    CURRENCY_FORMS = {}
    CURRENCY_ADJECTIVES = {}

    def __init__(self):
        self.is_title = False
        self.precision = 2
        self.exclude_title = []
        self.negword = "(-) "
        self.pointword = "(.)"
        self.errmsg_nonnum = "type(%s) not in [long, int, float]"
        self.errmsg_floatord = "Cannot treat float %s as ordinal."
        self.errmsg_negord = "Cannot treat negative num %s as ordinal."
        self.errmsg_toobig = "abs(%s) must be less than %s."

        self.setup()

        # uses cards
        if any(hasattr(self, field) for field in
               ['high_numwords', 'mid_numwords', 'low_numwords']):
            self.cards = OrderedDict()
            self.set_numwords()
            self.MAXVAL = 1000 * list(self.cards.keys())[0]

    def set_numwords(self):
        self.set_high_numwords(self.high_numwords)
        self.set_mid_numwords(self.mid_numwords)
        self.set_low_numwords(self.low_numwords)

    def set_high_numwords(self, *args):
        raise NotImplementedError

    def set_mid_numwords(self, mid):
        for key, val in mid:
            self.cards[key] = val

    def set_low_numwords(self, numwords):
        for word, n in zip(numwords, range(len(numwords) - 1, -1, -1)):
            self.cards[n] = word

    def splitnum(self, value):
        for elem in self.cards:
            if elem > value:
                continue

            out = []
            if value == 0:
                div, mod = 1, 0
            else:
                div, mod = divmod(value, elem)

            if div == 1:
                out.append((self.cards[1], 1))
            else:
                if div == value:  # The system tallies, eg Roman Numerals
                    return [(div * self.cards[elem], div*elem)]
                out.append(self.splitnum(div))

            out.append((self.cards[elem], elem))

            if mod:
                out.append(self.splitnum(mod))

            return out

    def parse_minus(self, num_str):
        """Detach minus and return it as symbol with new num_str."""
        if num_str.startswith('-'):
            # Extra spacing to compensate if there is no minus.
            return '%s ' % self.negword.strip(), num_str[1:]
        return '', num_str

    def str_to_number(self, value):
        return Decimal(value)

    def to_cardinal(self, value):
        try:
            assert int(value) == value
        except (ValueError, TypeError, AssertionError):
            return self.to_cardinal_float(value)

        out = ""
        if value < 0:
            value = abs(value)
            out = "%s " % self.negword.strip()

        if value >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (value, self.MAXVAL))

        val = self.splitnum(value)
        words, num = self.clean(val)
        return self.title(out + words)

    def float2tuple(self, value):
        pre = int(value)

        # Simple way of finding decimal places to update the precision
        self.precision = abs(Decimal(str(value)).as_tuple().exponent)

        post = abs(value - pre) * 10**self.precision
        if abs(round(post) - post) < 0.01:
            # We generally floor all values beyond our precision (rather than
            # rounding), but in cases where we have something like 1.239999999,
            # which is probably due to python's handling of floats, we actually
            # want to consider it as 1.24 instead of 1.23
            post = int(round(post))
        else:
            post = int(math.floor(post))

        return pre, post

    def to_cardinal_float(self, value):
        try:
            float(value) == value
        except (ValueError, TypeError, AssertionError, AttributeError):
            raise TypeError(self.errmsg_nonnum % value)

        pre, post = self.float2tuple(float(value))

        post = str(post)
        post = '0' * (self.precision - len(post)) + post

        out = [self.to_cardinal(pre)]
        if self.precision:
            out.append(self.title(self.pointword))

        for i in range(self.precision):
            curr = int(post[i])
            out.append(to_s(self.to_cardinal(curr)))

        return " ".join(out)

    def merge(self, curr, next):
        raise NotImplementedError

    def clean(self, val):
        out = val
        while len(val) != 1:
            out = []
            left, right = val[:2]
            if isinstance(left, tuple) and isinstance(right, tuple):
                out.append(self.merge(left, right))
                if val[2:]:
                    out.append(val[2:])
            else:
                for elem in val:
                    if isinstance(elem, list):
                        if len(elem) == 1:
                            out.append(elem[0])
                        else:
                            out.append(self.clean(elem))
                    else:
                        out.append(elem)
            val = out
        return out[0]

    def title(self, value):
        if self.is_title:
            out = []
            value = value.split()
            for word in value:
                if word in self.exclude_title:
                    out.append(word)
                else:
                    out.append(word[0].upper() + word[1:])
            value = " ".join(out)
        return value

    def verify_ordinal(self, value):
        if not value == int(value):
            raise TypeError(self.errmsg_floatord % value)
        if not abs(value) == value:
            raise TypeError(self.errmsg_negord % value)

    def to_ordinal(self, value):
        return self.to_cardinal(value)

    def to_ordinal_num(self, value):
        return value

    # Trivial version
    def inflect(self, value, text):
        text = text.split("/")
        if value == 1:
            return text[0]
        return "".join(text)

    # //CHECK: generalise? Any others like pounds/shillings/pence?
    def to_splitnum(self, val, hightxt="", lowtxt="", jointxt="",
                    divisor=100, longval=True, cents=True):
        out = []

        if isinstance(val, float):
            high, low = self.float2tuple(val)
        else:
            try:
                high, low = val
            except TypeError:
                high, low = divmod(val, divisor)

        if high:
            hightxt = self.title(self.inflect(high, hightxt))
            out.append(self.to_cardinal(high))
            if low:
                if longval:
                    if hightxt:
                        out.append(hightxt)
                    if jointxt:
                        out.append(self.title(jointxt))
            elif hightxt:
                out.append(hightxt)

        if low:
            if cents:
                out.append(self.to_cardinal(low))
            else:
                out.append("%02d" % low)
            if lowtxt and longval:
                out.append(self.title(self.inflect(low, lowtxt)))

        return " ".join(out)

    def to_year(self, value, **kwargs):
        return self.to_cardinal(value)

    def pluralize(self, n, forms):
        """
        Should resolve gettext form:
        http://docs.translatehouse.org/projects/localization-guide/en/latest/l10n/pluralforms.html
        """
        raise NotImplementedError

    def _money_verbose(self, number, currency):
        return self.to_cardinal(number)

    def _cents_verbose(self, number, currency):
        return self.to_cardinal(number)

    def _cents_terse(self, number, currency):
        return "%02d" % number

    def to_currency(self, val, currency='EUR', cents=True, separator=',',
                    adjective=False):
        """
        Args:
            val: Numeric value
            currency (str): Currency code
            cents (bool): Verbose cents
            separator (str): Cent separator
            adjective (bool): Prefix currency name with adjective
        Returns:
            str: Formatted string

        """
        left, right, is_negative = parse_currency_parts(val)

        try:
            cr1, cr2 = self.CURRENCY_FORMS[currency]

        except KeyError:
            raise NotImplementedError(
                'Currency code "%s" not implemented for "%s"' %
                (currency, self.__class__.__name__))

        if adjective and currency in self.CURRENCY_ADJECTIVES:
            cr1 = prefix_currency(self.CURRENCY_ADJECTIVES[currency], cr1)

        minus_str = "%s " % self.negword.strip() if is_negative else ""
        money_str = self._money_verbose(left, currency)
        cents_str = self._cents_verbose(right, currency) \
            if cents else self._cents_terse(right, currency)

        return u'%s%s %s%s %s %s' % (
            minus_str,
            money_str,
            self.pluralize(left, cr1),
            separator,
            cents_str,
            self.pluralize(right, cr2)
        )

    def setup(self):
        pass



# coding: utf-8
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from unittest import TestCase

from num2words import num2words


class Num2WordsSVTest(TestCase):
    def test_ordinal(self):
        self.assertEqual(num2words(14, to="ordinal", lang="sv"), "fjortonde")
        self.assertEqual(num2words(1435, to="ordinal", lang="sv"),
                         "etttusen fyrahundratrettiofemte")
        self.assertEqual(num2words(32, to="ordinal", lang="sv"),
                         "trettioandra")
        self.assertEqual(num2words(1, to="ordinal", lang="sv"), "första")
        self.assertEqual(num2words(5, to="ordinal", lang="sv"), "femte")
        self.assertEqual(num2words(10, to="ordinal", lang="sv"), "tionde")

    def test_cardinal(self):
        self.assertEqual(num2words(0, to="cardinal", lang="sv"), "noll")
        self.assertEqual(num2words(1, to="cardinal", lang="sv"), "ett")
        self.assertEqual(num2words(3, to="cardinal", lang="sv"), "tre")
        self.assertEqual(num2words(5, to="cardinal", lang="sv"), "fem")
        self.assertEqual(num2words(18, to="cardinal", lang="sv"), "arton")
        self.assertEqual(num2words(45, to="cardinal", lang="sv"), "förtiofem")
        self.assertEqual(num2words(1345, to="cardinal", lang="sv"),
                         "etttusen trehundraförtiofem")
        self.assertEqual(num2words(4435, to="cardinal", lang="sv"),
                         "fyratusen fyrahundratrettiofem")
        self.assertEqual(num2words(1004135, to="cardinal", lang="sv"),
                         "en miljon fyratusen etthundratrettiofem")
        self.assertEqual(num2words(4335000, to="cardinal", lang="sv"),
                         "fyra miljoner trehundratrettiofemtusen")
        self.assertEqual(num2words(14004535, to="cardinal", lang="sv"),
                         "fjorton miljoner fyratusen femhundratrettiofem")
        self.assertEqual(num2words(1.5, to="cardinal", lang="sv"),
                         "ett komma fem")

    def test_not_implemented_options(self):
        with self.assertRaises(NotImplementedError) as context:
            num2words(1235, to="year", lang="sv")
        self.assertTrue("'year' is not implemented for swedish language"
                        in str(context.exception))

        with self.assertRaises(NotImplementedError) as context:
            num2words(1235, to="currency", lang="sv")
        self.assertTrue("'currency' is not implemented for swedish language"
                        in str(context.exception))

        with self.assertRaises(NotImplementedError) as context:
            num2words(1235, to="ordinal_num", lang="sv")
        self.assertTrue("'ordinal_num' is not implemented for swedish language"
                        in str(context.exception))



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA
from __future__ import division, unicode_literals

import re

from . import lang_PT


class Num2Word_PT_BR(lang_PT.Num2Word_PT):
    def set_high_numwords(self, high):
        max = 3 + 3*len(high)
        for word, n in zip(high, range(max, 3, -3)):
            self.cards[10**n] = word + "ilhão"

    def setup(self):
        super(Num2Word_PT_BR, self).setup()

        self.low_numwords[1] = 'dezenove'
        self.low_numwords[3] = 'dezessete'
        self.low_numwords[4] = 'dezesseis'

        self.thousand_separators = {
            3: "milésimo",
            6: "milionésimo",
            9: "bilionésimo",
            12: "trilionésimo",
            15: "quadrilionésimo"
        }

    def merge(self, curr, next):
        ctext, cnum, ntext, nnum = curr + next

        if cnum == 1:
            if nnum < 1000000:
                return next
            ctext = "um"
        elif cnum == 100 and nnum % 1000 != 0:
            ctext = "cento"

        if nnum < cnum:
            if cnum < 100:
                return ("%s e %s" % (ctext, ntext), cnum + nnum)
            return ("%s e %s" % (ctext, ntext), cnum + nnum)

        elif (not nnum % 1000000) and cnum > 1:
            ntext = ntext[:-4] + "lhões"

        if nnum == 100:
            ctext = self.hundreds[cnum]
            ntext = ""

        else:
            ntext = " " + ntext

        return (ctext + ntext, cnum * nnum)

    def to_cardinal(self, value):
        result = lang_PT.Num2Word_EU.to_cardinal(self, value)

        # Transforms "mil E cento e catorze reais" into "mil, cento e catorze
        # reais"
        for ext in (
                'mil', 'milhão', 'milhões', 'bilhão', 'bilhões',
                'trilhão', 'trilhões', 'quatrilhão', 'quatrilhões'):
            if re.match('.*{} e \\w*entos? (?=.*e)'.format(ext), result):
                result = result.replace(
                    '{} e'.format(ext), '{},'.format(ext), 1
                )

        return result

    def to_currency(self, val, longval=True):
        integer_part, decimal_part = ('%.2f' % val).split('.')

        result = self.to_cardinal(int(integer_part))

        appended_currency = False
        for ext in (
                'milhão', 'milhões', 'bilhão', 'bilhões',
                'trilhão', 'trilhões', 'quatrilhão', 'quatrilhões'):
            if result.endswith(ext):
                result += ' de reais'
                appended_currency = True

        if result in ['um', 'menos um']:
            result += ' real'
            appended_currency = True
        if not appended_currency:
            result += ' reais'

        if int(decimal_part):
            cents = self.to_cardinal(int(decimal_part))
            result += ' e ' + cents

            if cents == 'um':
                result += ' centavo'
            else:
                result += ' centavos'

        return result



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from decimal import Decimal
from unittest import TestCase

from num2words.currency import parse_currency_parts


class CurrencyTestCase(TestCase):
    def test_parse_currency_parts(self):
        # integer with cents
        self.assertEqual(parse_currency_parts(101), (1, 1, False))
        self.assertEqual(parse_currency_parts(-123), (1, 23, True))

        # integer without cents
        self.assertEqual(parse_currency_parts(101, is_int_with_cents=False),
                         (101, 0, False))
        self.assertEqual(parse_currency_parts(-123, is_int_with_cents=False),
                         (123, 0, True))

        # float
        self.assertEqual(parse_currency_parts(1.01), (1, 1, False))
        self.assertEqual(parse_currency_parts(-1.23), (1, 23, True))
        self.assertEqual(parse_currency_parts(-1.2), (1, 20, True))
        self.assertEqual(parse_currency_parts(0.004), (0, 0, False))
        self.assertEqual(parse_currency_parts(0.005), (0, 1, False))
        self.assertEqual(parse_currency_parts(0.006), (0, 1, False))
        self.assertEqual(parse_currency_parts(0.0005), (0, 0, False))
        self.assertEqual(parse_currency_parts(0.984), (0, 98, False))
        self.assertEqual(parse_currency_parts(0.989), (0, 99, False))
        self.assertEqual(parse_currency_parts(0.994), (0, 99, False))
        self.assertEqual(parse_currency_parts(0.999), (1, 0, False))
        # self.assertEqual(parse_currency_parts(0.985), (0, 99, False))
        # self.assertEqual(parse_currency_parts(0.995), (1, 0, False))

        # decimal
        self.assertEqual(parse_currency_parts(Decimal("1.01")), (1, 1, False))
        self.assertEqual(parse_currency_parts(Decimal("-1.23")), (1, 23, True))
        self.assertEqual(parse_currency_parts(Decimal("-1.233")),
                         (1, 23, True))
        self.assertEqual(parse_currency_parts(Decimal("-1.989")),
                         (1, 99, True))

        # string
        self.assertEqual(parse_currency_parts("1.01"), (1, 1, False))
        self.assertEqual(parse_currency_parts("-1.23"), (1, 23, True))
        self.assertEqual(parse_currency_parts("-1.2"), (1, 20, True))
        self.assertEqual(parse_currency_parts("1"), (1, 0, False))



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from unittest import TestCase

from num2words import num2words


class Num2WordsROTest(TestCase):

    def test_ordinal(self):
        self.assertEqual(
            num2words(1, lang='ro', to='ordinal'),
            'primul'
        )
        self.assertEqual(
            num2words(22, lang='ro', to='ordinal'),
            u'al douăzeci și doilea'
        )
        self.assertEqual(
            num2words(21, lang='ro', to='ordinal'),
            u'al douăzeci și unulea'
        )
        self.assertEqual(
            num2words(12, lang='ro', to='ordinal'),
            u'al doisprezecelea'
        )
        self.assertEqual(
            num2words(130, lang='ro', to='ordinal'),
            u'al o sută treizecilea'
        )
        self.assertEqual(
            num2words(1003, lang='ro', to='ordinal'),
            u'al o mie treilea'
        )

    def test_ordinal_num(self):
        self.assertEqual(
            num2words(1, lang='ro', to='ordinal_num'),
            '1-ul'
        )
        self.assertEqual(
            num2words(10, lang='ro', to='ordinal_num'),
            'al 10-lea'
        )
        self.assertEqual(num2words(
            21, lang='ro', to='ordinal_num'),
            'al 21-lea'
        )
        self.assertEqual(
            num2words(102, lang='ro', to='ordinal_num'),
            'al 102-lea'
        )
        self.assertEqual(
            num2words(73, lang='ro', to='ordinal_num'),
            'al 73-lea'
        )

    def test_cardinal_for_float_number(self):
        self.assertEqual(
            num2words(12.5, lang='ro'),
            u'doisprezece virgulă cinci'
        )
        self.assertEqual(
            num2words(12.51, lang='ro'),
            u'doisprezece virgulă cinci unu'
        )
        self.assertEqual(
            num2words(12.53, lang='ro'),
            u'doisprezece virgulă cinci trei'
        )
        self.assertEqual(
            num2words(12.59, lang='ro'),
            u'doisprezece virgulă cinci nouă'
        )

    def test_big_numbers(self):
        self.assertEqual(
            num2words(1000000, lang="ro"),
            u"un milion"
        )
        self.assertEqual(
            num2words(1000000000, lang="ro"),
            u"un miliard"
        )
        self.assertEqual(
            num2words(33000000, lang="ro"),
            u"treizeci și trei milioane"
        )
        self.assertEqual(
            num2words(247000000000, lang="ro"),
            u"două sute patruzeci și șapte de miliarde"
        )

    def test_overflow(self):
        with self.assertRaises(OverflowError):
            num2words("100000000000000000000000000000000000000000000000000000"
                      "000000000000000000000000000000000000000000000000000000"
                      "000000000000000000000000000000000000000000000000000000"
                      "000000000000000000000000000000000000000000000000000000"
                      "000000000000000000000000000000000000000000000000000000"
                      "0000000000000000000000000000000000000")

    def test_to_currency(self):
        self.assertEqual(
            num2words(1000, lang='ro', to='currency'),
            u'una mie de lei'
        )
        self.assertEqual(
            num2words(101, lang='ro', to='currency'),
            u'una sută unu lei'
        )
        self.assertEqual(
            num2words(100, lang='ro', to='currency'),
            u'una sută de lei'
        )
        self.assertEqual(
            num2words(38.4, lang='ro', to='currency'),
            u'treizeci și opt de lei și patruzeci de bani'
        )
        self.assertEqual(
            num2words(1.01, lang='ro', to='currency'),
            u'un leu și un ban'
        )
        self.assertEqual(
            num2words(4778.00, lang='ro', to='currency'),
            u'patru mii șapte sute șaptezeci și opt de lei')
        self.assertEqual(
            num2words(4778.32, lang='ro', to='currency'),
            u'patru mii șapte sute șaptezeci și opt de lei'
            u' și treizeci și doi de bani')
        self.assertEqual(
            num2words(1207, lang='ro', to='currency'),
            u'una mie două sute șapte lei')
        self.assertEqual(
            num2words(22000, lang='ro', to='currency'),
            u'douăzeci și două de mii de lei')
        self.assertEqual(
            num2words(80000, lang='ro', to='currency'),
            u'optzeci de mii de lei')
        self.assertEqual(
            num2words(123456789, lang='ro', to='currency'),
            u'una sută douăzeci și trei milioane patru sute '
            u'cincizeci și șase de mii șapte sute optzeci și nouă de lei')

    def test_to_year(self):
        self.assertEqual(num2words(1989, lang='ro', to='year'),
                         u'o mie nouă sute optzeci și nouă')
        self.assertEqual(num2words(1984, lang='ro', to='year'),
                         u'o mie nouă sute optzeci și patru')
        self.assertEqual(num2words(2018, lang='ro', to='year'),
                         u'două mii optsprezece')
        self.assertEqual(num2words(1066, lang='ro', to='year'),
                         u'o mie șaizeci și șase')
        self.assertEqual(num2words(5000, lang='ro', to='year'),
                         u'cinci mii')
        self.assertEqual(num2words(2001, lang='ro', to='year'),
                         u'două mii unu')
        self.assertEqual(num2words(905, lang='ro', to='year'),
                         u'nouă sute cinci')
        self.assertEqual(num2words(6600, lang='ro', to='year'),
                         u'șase mii șase sute')
        self.assertEqual(num2words(1600, lang='ro', to='year'),
                         u'o mie șase sute')
        self.assertEqual(num2words(700, lang='ro', to='year'),
                         u'șapte sute')
        self.assertEqual(num2words(50, lang='ro', to='year'),
                         u'cincizeci')
        self.assertEqual(num2words(0, lang='ro', to='year'),
                         u'zero')
        self.assertEqual(num2words(10, lang='ro', to='year'),
                         u'zece')
        # suffixes
        self.assertEqual(num2words(-44, lang='ro', to='year'),
                         u'patruzeci și patru î.Hr.')
        self.assertEqual(num2words(-44, lang='ro', to='year',
                         suffix=u'î.e.n.'), u'patruzeci și patru î.e.n.')
        self.assertEqual(num2words(1, lang='ro', to='year', suffix='d.Hr.'),
                         u'unu d.Hr.')
        self.assertEqual(num2words(-66000000, lang='ro', to='year'),
                         u'șaizeci și șase milioane î.Hr.')



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA


from __future__ import print_function, unicode_literals

from .base import Num2Word_Base
from .compat import to_s
from .currency import parse_currency_parts
from .utils import get_digits, splitbyx

ZERO = (u'אפס',)

ONES = {
    1: (u'אחת', u'אחד', u'אחת', u'אחד',
        u'ראשונה', u'ראשון', u'ראשונות', u'ראשונים'),
    2: (u'שתיים', u'שניים', u'שתי', u'שני',
        u'שנייה', u'שני', u'שניות', u'שניים'),
    3: (u'שלוש', u'שלושה', u'שלוש', u'שלושת',
        u'שלישית', u'שלישי', u'שלישיות', u'שלישיים'),
    4: (u'ארבע', u'ארבעה', u'ארבע', u'ארבעת',
        u'רביעית', u'רביעי', u'רביעיות', u'רביעיים'),
    5: (u'חמש', u'חמישה', u'חמש', u'חמשת',
        u'חמישית', u'חמישי', u'חמישיות', u'חמישיים'),
    6: (u'שש', u'שישה', u'שש', u'ששת',
        u'שישית', u'שישי', u'שישיות', u'שישיים'),
    7: (u'שבע', u'שבעה', u'שבע', u'שבעת',
        u'שביעית', u'שביעי', u'שביעיות', u'שביעיים'),
    8: (u'שמונה', u'שמונה', u'שמונה', u'שמונת',
        u'שמינית', u'שמיני', u'שמיניות', u'שמיניים'),
    9: (u'תשע', u'תשעה', u'תשע', u'תשעת',
        u'תשיעית', u'תשיעי', u'תשיעיות', u'תשיעיים'),
}

TENS = {
    0: (u'עשר', u'עשרה', u'עשר', u'עשרת',
        u'עשירית', u'עשירי', u'עשיריות', u'עשיריים'),
    1: (u'עשרה', u'עשר'),
    2: (u'שתים עשרה', u'שנים עשר'),
}

TWENTIES = {
    2: (u'עשרים',),
    3: (u'שלושים',),
    4: (u'ארבעים',),
    5: (u'חמישים',),
    6: (u'שישים',),
    7: (u'שבעים',),
    8: (u'שמונים',),
    9: (u'תשעים',),
}

HUNDREDS = {
    1: (u'מאה', u'מאת'),
    2: (u'מאתיים',),
    3: (u'מאות',)
}

THOUSANDS = {
    1: (u'אלף',),
    2: (u'אלפיים',),
    3: (u'אלפים', 'אלפי'),
}

LARGE = {
    1: (u'מיליון', u'מיליוני'),
    2: (u'מיליארד', u'מיליארדי'),
    3: (u'טריליון', u'טריליוני'),
    4: (u'קוודריליון', u'קוודריליוני'),
    5: (u'קווינטיליון', u'קווינטיליוני'),
    6: (u'סקסטיליון', u'סקסטיליוני'),
    7: (u'ספטיליון', u'ספטיליוני'),
    8: (u'אוקטיליון', u'אוקטיליוני'),
    9: (u'נוניליון', u'נוניליוני'),
    10: (u'דסיליון', u'דסיליוני'),
    11: (u'אונדסיליון', u'אונדסיליוני'),
    12: (u'דואודסיליון', u'דואודסיליוני'),
    13: (u'טרדסיליון', u'טרדסיליוני'),
    14: (u'קווטואורדסיליון', u'קווטואורדסיליוני'),
    15: (u'קווינדסיליון', u'קווינדסיליוני'),
    16: (u'סקסדסיליון', u'סקסדסיליוני'),
    17: (u'ספטנדסיליון', u'ספטנדסיליוני'),
    18: (u'אוקטודסיליון', u'אוקטודסיליוני'),
    19: (u'נובמדסיליון', u'נובמדסיליוני'),
    20: (u'ויגינטיליון', u'ויגינטיליוני')
}

AND = u'ו'

DEF = u'ה'

MAXVAL = int('1' + '0'*66)


def chunk2word(n, i, x, gender='f', construct=False,
               ordinal=False, plural=False):
    words = []
    n1, n2, n3 = get_digits(x)

    if n3 > 0:
        if construct and n == 100:
            words.append(HUNDREDS[n3][1])
        elif n3 <= 2:
            words.append(HUNDREDS[n3][0])
        else:
            words.append(ONES[n3][0] + ' ' + HUNDREDS[3][0])

    if n2 > 1:
        words.append(TWENTIES[n2][0])

    if i == 0 or x >= 11:
        male = gender == 'm' or i > 0
        cop = (2*(construct and i == 0)+4*ordinal+2*plural) * (n < 11)
        if n2 == 1:
            if n1 == 0:
                words.append(TENS[n1][male + cop])
            elif n1 == 2:
                words.append(TENS[n1][male])
            else:
                words.append(ONES[n1][male] + ' ' + TENS[1][male])
        elif n1 > 0:
            words.append(ONES[n1][male + cop])

    construct_last = construct and (n % 1000 ** i == 0)

    if i == 1:
        if x >= 11:
            words[-1] = words[-1] + ' ' + THOUSANDS[1][0]
        elif n1 == 0:
            words.append(TENS[0][3] + ' ' + THOUSANDS[3][construct_last])
        elif n1 <= 2:
            words.append(THOUSANDS[n1][0])
        else:
            words.append(ONES[n1][3] + ' ' + THOUSANDS[3][construct_last])

    elif i > 1:
        if x >= 11:
            words[-1] = words[-1] + ' ' + LARGE[i - 1][construct_last]
        elif n1 == 0:
            words.append(TENS[0][1 + 2*construct_last] + ' ' +
                         LARGE[i - 1][construct_last])
        elif n1 == 1:
            words.append(LARGE[i - 1][0])
        else:
            words.append(ONES[n1][1 + 2*(construct_last or x == 2)] + ' ' +
                         LARGE[i - 1][construct_last])

    return words


def int2word(n, gender='f', construct=False,
             ordinal=False, definite=False, plural=False):
    assert n == int(n)
    assert not construct or not ordinal
    assert ordinal or (not definite and not plural)
    if n >= MAXVAL:
        raise OverflowError('abs(%s) must be less than %s.' % (n, MAXVAL))

    if n == 0:
        if ordinal:
            return DEF + ZERO[0]
        return ZERO[0]

    words = []

    chunks = list(splitbyx(str(n), 3))
    i = len(chunks)
    for x in chunks:
        i -= 1

        if x == 0:
            continue

        words += chunk2word(n, i, x, gender=gender, construct=construct,
                            ordinal=ordinal, plural=plural)

        # https://hebrew-academy.org.il/2017/01/30/%D7%95-%D7%94%D7%97%D7%99%D7%91%D7%95%D7%A8-%D7%91%D7%9E%D7%A1%D7%A4%D7%A8%D7%99%D7%9D  # noqa
        if len(words) > 1:
            words[-1] = AND + words[-1]

    if ordinal and (n >= 11 or definite):
        words[0] = DEF + words[0]

    return ' '.join(words)


class Num2Word_HE(Num2Word_Base):
    CURRENCY_FORMS = {
        'ILS': ((u'שקל', u'שקלים'), (u'אגורה', u'אגורות')),
        'EUR': ((u'אירו', u'אירו'), (u'סנט', u'סנטים')),
        'USD': ((u'דולר', u'דולרים'), (u'סנט', u'סנטים')),
    }

    CURRENCY_GENDERS = {
        'ILS': ('m', 'f'),
        'EUR': ('m', 'm'),
        'USD': ('m', 'm'),
    }

    def __init__(self, makaf='-'):
        super(Num2Word_HE, self).__init__()
        self.makaf = makaf

    def setup(self):
        super(Num2Word_HE, self).setup()
        self.negword = u'מינוס'
        self.pointword = u'נקודה'
        self.MAXVAL = MAXVAL

    def to_cardinal_float(self, value, gender='f'):
        try:
            float(value) == value
        except (ValueError, TypeError, AssertionError, AttributeError):
            raise TypeError(self.errmsg_nonnum % value)

        pre, post = self.float2tuple(float(value))

        post = str(post)
        post = '0'*(self.precision - len(post)) + post

        out = [self.to_cardinal(pre, gender=gender)]
        if self.precision:
            out.append(self.title(self.pointword))

        for i in range(self.precision):
            curr = int(post[i])
            out.append(to_s(self.to_cardinal(curr)))

        return ' '.join(out)

    def to_cardinal(self, value, gender='f', construct=False):
        try:
            assert int(value) == value
        except (ValueError, TypeError, AssertionError):
            # https://hebrew-academy.org.il/2019/12/03/%D7%A2%D7%9C-%D7%94%D7%91%D7%A2%D7%AA-%D7%94%D7%9E%D7%A1%D7%A4%D7%A8-%D7%94%D7%9E%D7%A2%D7%95%D7%A8%D7%91  # noqa
            return self.to_cardinal_float(value, gender=gender)

        out = ""
        if value < 0:
            value = abs(value)
            out = "%s " % self.negword.strip()

        if value >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (value, self.MAXVAL))

        return out + int2word(int(value), gender=gender, construct=construct)

    def to_ordinal(self, value, gender='m', definite=False, plural=False):
        self.verify_ordinal(value)

        if value >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (value, self.MAXVAL))

        return int2word(int(value), gender=gender, ordinal=True,
                        definite=definite, plural=plural)

    def pluralize(self, n, forms, currency=None, prefer_singular=False):
        assert n == int(n)
        form = 1
        if n == 1 or prefer_singular and (
                abs(n) >= 11 or n == 0 or currency != 'ILS'):
            form = 0
        return forms[form]

    def to_currency(self, val, currency='ILS', cents=True,
                    separator=AND, adjective=False,
                    prefer_singular=False, prefer_singular_cents=False):
        left, right, is_negative = parse_currency_parts(val)

        if not separator.startswith(' '):
            separator = ' ' + separator

        try:
            cr1, cr2 = self.CURRENCY_FORMS[currency]

        except KeyError:
            raise NotImplementedError(
                'Currency code "%s" not implemented for "%s"' %
                (currency, self.__class__.__name__))

        minus_str = "%s " % self.negword.strip() if is_negative else ""
        try:
            gender1, gender2 = self.CURRENCY_GENDERS[currency]
        except KeyError:
            gender1 = gender2 = ''

        money_str = self.to_cardinal(left, gender=gender1,
                                     construct=left == 2)
        if cents:
            cents_str = self.to_cardinal(right, gender=gender2,
                                         construct=right == 2)
        else:
            cents_str = self._cents_terse(right, currency)
            sep_parts = separator.split()
            if sep_parts and sep_parts[-1] == AND:
                separator += self.makaf

        strings = [
            minus_str,
            money_str,
            self.pluralize(left, cr1, currency=currency,
                           prefer_singular=prefer_singular),
            separator,
            cents_str,
            self.pluralize(right, cr2, currency=currency,
                           prefer_singular=prefer_singular_cents)
        ]
        if left == 1:
            strings[1], strings[2] = strings[2], strings[1]
        if right == 1:
            strings[4], strings[5] = strings[5], strings[4]
        # In Hebrew the separator is along with the following word
        return u'%s%s %s%s%s %s' % tuple(strings)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from .lang_EU import Num2Word_EU


class Num2Word_TE(Num2Word_EU):
    def set_high_numwords(self, high):
        for n, word in self.high_numwords:
            self.cards[10**n] = word

    def setup(self):
        self.low_numwords = [
            "తొంభై తొమ్మిది",
            "తొంభై ఎనిమిది",
            "తొంభై ఏడు",
            "తొంభై ఆరు",
            "తొంభై అయిదు",
            "తొంభై నాలుగు",
            "తొంభై మూడు",
            "తొంభై రెండు",
            "తొంభై ఒకటి",
            "తొంభై",
            "ఎనభై తొమ్మిది",
            "ఎనభై ఎనిమిది",
            "ఎనభై ఏడు",
            "ఎనభై ఆరు",
            "ఎనభై అయిదు",
            "ఎనభై నాలుగు",
            "ఎనభై మూడు",
            "ఎనభై రెండు",
            "ఎనభై ఒకటి",
            "ఎనభై",
            "డెబ్బై తొమ్మిది",
            "డెబ్బై ఎనిమిది",
            "డెబ్బై ఏడు",
            "డెబ్బై ఆరు",
            "డెబ్బై అయిదు",
            "డెబ్బై నాలుగు",
            "డెబ్బై మూడు",
            "డెబ్బై రెండు",
            "డెబ్బై ఒకటి",
            "డెబ్బై",
            "అరవై తొమ్మిది",
            "అరవై ఎనిమిది",
            "అరవై ఏడు",
            "అరవై ఆరు",
            "అరవై అయిదు",
            "అరవై నాలుగు",
            "అరవై మూడు",
            "అరవై రెండు",
            "అరవై ఒకటి",
            "అరవై",
            "యాభై తొమ్మిది",
            "యాభై ఎనిమిది",
            "యాభై ఏడు",
            "యాభై ఆరు",
            "యాభై అయిదు",
            "యాభై నాలుగు",
            "యాభై మూడు",
            "యాభై రెండు",
            "యాభై ఒకటి",
            "యాభై ",
            "నలభై తొమ్మిది",
            "నలభై ఎనిమిది",
            "నలభై ఏడు",
            "నలభై ఆరు",
            "నలభై అయిదు",
            "నలభై నాలుగు",
            "నలభై మూడు",
            "నలభై రెండు",
            "నలభై ఒకటి",
            "నలభై",
            "ముప్పై తొమ్మిది",
            "ముప్పై ఎనిమిది",
            "ముప్పై ఏడు",
            "ముప్పై ఆరు",
            "ముప్పై ఐదు",
            "ముప్పై నాలుగు",
            "ముప్పై మూడు",
            "ముప్పై రెండు",
            "ముప్పై ఒకటి",
            "ముప్పై",
            "ఇరవై తొమ్మిది",
            "ఇరవై ఎనిమిది",
            "ఇరవై ఏడు",
            "ఇరవై ఆరు",
            "ఇరవై అయిదు",
            "ఇరవై నాలుగు",
            "ఇరవై మూడు",
            "ఇరవై రెండు",
            "ఇరవై ఒకటి",
            "ఇరవై",
            "పందొమ్మిది",
            "పధ్ధెనిమిది",
            "పదిహేడు",
            "పదహారు",
            "పదునయిదు",
            "పధ్నాలుగు",
            "పదమూడు",
            "పన్నెండు",
            "పదకొండు",
            "పది",
            "తొమ్మిది",
            "ఎనిమిది",
            "ఏడు",
            "ఆరు",
            "అయిదు",
            "నాలుగు",
            "మూడు",
            "రెండు",
            "ఒకటి",
            "సున్న",
        ]

        self.mid_numwords = [(100, "వంద")]

        self.high_numwords = [(7, "కోట్ల"), (5, "లక్ష"), (3, "వేయి")]

        self.pointword = "బిందువు "

        self.modifiers = [
            " ్  ",
            "ా ",
            " ి ",
            " ీ ",
            " ు ",
            " ూ ",
            " ృ ",
            " ౄ  ",
            " ె ",
            " ే ",
            " ై ",
            " ొ",
            " ో ",
            " ౌ ",
            " ఁ ",
            " ం ",
            " ః ",
        ]

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair
        if lnum == 1 and rnum < 100:
            return (rtext, rnum)
        elif 100 > lnum > rnum:
            return ("%s-%s" % (ltext, rtext), lnum + rnum)
        elif lnum >= 100 > rnum:
            if ltext[-1] in self.modifiers:
                return ("%s %s" % (ltext[:-1], rtext), lnum + rnum)
            else:
                return ("%s %s" % (ltext+"ల", rtext), lnum + rnum)
        elif rnum > lnum:
            return ("%s %s" % (ltext, rtext), lnum * rnum)
        return ("%s %s" % (ltext, rtext), lnum + rnum)

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return "%s%s" % (value, self.to_ordinal(value)[-1:])

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        outwords = self.to_cardinal(value)
        if outwords[-1] in self.modifiers:
            outwords = outwords[:-1]
        ordinal_num = outwords + "వ"
        return ordinal_num



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from unittest import TestCase

from num2words import num2words


class Num2WordsKZTest(TestCase):
    def test_to_cardinal(self):
        self.maxDiff = None
        self.assertEqual(num2words(7, lang="kz"), "жеті")
        self.assertEqual(num2words(23, lang="kz"), "жиырма үш")
        self.assertEqual(num2words(145, lang="kz"), "жүз қырық бес")
        self.assertEqual(
            num2words(2869, lang="kz"),
            "екі мың сегіз жүз алпыс тоғыз"
        )
        self.assertEqual(
            num2words(-789000125, lang="kz"),
            "минус жеті жүз сексен тоғыз миллион жүз жиырма бес",
        )
        self.assertEqual(
            num2words(84932, lang="kz"), "сексен төрт мың тоғыз жүз отыз екі"
        )

    def test_to_cardinal_floats(self):
        self.assertEqual(num2words(100.67, lang="kz"), "жүз бүтін алпыс жеті")
        self.assertEqual(num2words(0.7, lang="kz"), "нөл бүтін жеті")
        self.assertEqual(num2words(1.73, lang="kz"), "бір бүтін жетпіс үш")
        self.assertEqual(
            num2words(10.02, lang='kz'),
            "он бүтін нөл екі"
        )
        self.assertEqual(
            num2words(15.007, lang='kz'),
            "он бес бүтін нөл нөл жеті"
        )

    def test_to_ordinal(self):
        with self.assertRaises(NotImplementedError):
            num2words(1, lang="kz", to="ordinal")

    def test_to_currency(self):
        self.assertEqual(
            num2words(25.24, lang="kz", to="currency", currency="KZT"),
            "жиырма бес теңге, жиырма төрт тиын",
        )
        self.assertEqual(
            num2words(1996.4, lang="kz", to="currency", currency="KZT"),
            "бір мың тоғыз жүз тоқсан алты теңге, қырық тиын",
        )
        self.assertEqual(
            num2words(632924.51, lang="kz", to="currency", currency="KZT"),
            "алты жүз отыз екі мың тоғыз жүз жиырма төрт теңге, елу бір тиын",
        )
        self.assertEqual(
            num2words(632924.513, lang="kz", to="currency", currency="KZT"),
            "алты жүз отыз екі мың тоғыз жүз жиырма төрт теңге, елу бір тиын",
        )
        self.assertEqual(
            num2words(987654321.123, lang="kz", to="currency", currency="KZT"),
            "тоғыз жүз сексен жеті миллион алты жүз елу төрт мың "
            "үш жүз жиырма бір теңге, он екі тиын",
        )



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from unittest import TestCase

from num2words import num2words

from . import test_fr

TEST_CASES_TO_CURRENCY = (
    (1, 'un dinard'),
    (2, 'deux dinards'),
    (8, 'huit dinards'),
    (12, 'douze dinards'),
    (21, 'vingt et un dinards'),
    (81.25, 'quatre-vingt-un dinards et vingt-cinq centimes'),
    (100, 'cent dinards'),
)


class Num2WordsPLTest(TestCase):
    def test_currency(self):
        self.assertEqual(
            num2words(1234.12, lang='fr_DZ', to='currency'),
            "mille deux cent trente-quatre dinards et douze centimes"
        )
        self.assertEqual(
            num2words(45689.89, lang='fr_DZ', to='currency'),
            "quarante-cinq mille six cent quatre-vingt-neuf dinards et "
            "quatre-vingt-neuf centimes"
        )

    def test_number(self):
        for test in test_fr.TEST_CASES_CARDINAL:
            self.assertEqual(num2words(test[0], lang='fr_DZ'), test[1])

    def test_ordinal(self):
        for test in test_fr.TEST_CASES_ORDINAL:
            self.assertEqual(
                num2words(test[0], lang='fr_DZ', ordinal=True),
                test[1]
            )

    def test_ordinal_num(self):
        for test in test_fr.TEST_CASES_ORDINAL_NUM:
            self.assertEqual(
                num2words(test[0], lang='fr_DZ', to='ordinal_num'),
                test[1]
            )



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from unittest import TestCase

from num2words import num2words


def n2k(*args, **kwargs):
    return num2words(*args, lang='ko', **kwargs)


class Num2WordsKOTest(TestCase):
    def test_low(self):
        cases = [(0, "영"), (1, "일"), (2, "이"), (3, "삼"), (4, "사"), (5, "오"),
                 (6, "육"), (7, "칠"), (8, "팔"), (9, "구"), (10, "십"),
                 (11, "십일"), (12, "십이"), (13, "십삼"), (14, "십사"),
                 (15, "십오"), (16, "십육"), (17, "십칠"),
                 (18, "십팔"), (19, "십구"), (20, "이십"), (25, "이십오"),
                 (31, "삼십일"), (42, "사십이"), (54, "오십사"), (63, "육십삼"),
                 (76, "칠십육"), (89, "팔십구"), (98, "구십팔")]
        for num, out in cases:
            self.assertEqual(n2k(num), out)

    def test_mid(self):
        cases = [(100, "백"), (121, "백이십일"), (160, "백육십"), (256, "이백오십육"),
                 (285, "이백팔십오"), (486, "사백팔십육"), (627, "육백이십칠"),
                 (808, "팔백팔"), (999, "구백구십구"), (1004, "천사"),
                 (2018, "이천십팔"), (7063, "칠천육십삼")]
        for num, out in cases:
            self.assertEqual(n2k(num), out)

    def test_high(self):
        cases = [(10000, "만"), (11020, "만 천이십"), (25891, "이만 오천팔백구십일"),
                 (64237, "육만 사천이백삼십칠"), (241572, "이십사만 천오백칠십이"),
                 (100000000, "일억"), (5000500000000, "오조 오억")]
        for num, out in cases:
            self.assertEqual(n2k(num), out)

    def test_negative(self):
        cases = [(-11, "마이너스 십일"), (-15, "마이너스 십오"),
                 (-18, "마이너스 십팔"), (-241572, "마이너스 이십사만 천오백칠십이")]
        for num, out in cases:
            self.assertEqual(n2k(num), out)

    def test_year(self):
        cases = [(2000, "이천년"), (2002, "이천이년"), (2018, "이천십팔년"),
                 (1954, "천구백오십사년"), (1910, "천구백십년"), (-1000, "기원전 천년")]
        for num, out in cases:
            self.assertEqual(n2k(num, to="year"), out)

    def test_currency(self):
        cases_krw = [(8350, "팔천삼백오십원"), (14980, "만사천구백팔십원"),
                     (250004000, "이억오천만사천원")]
        cases_usd = [(4, "사달러 영센트"), (19.55, "십구달러 오십오센트")]
        cases_jpy = [(15, "십오엔"), (50, "오십엔")]
        for num, out in cases_krw:
            self.assertEqual(n2k(num, to="currency"), out)
        for num, out in cases_usd:
            self.assertEqual(n2k(num, to="currency", currency="USD"), out)
        for num, out in cases_jpy:
            self.assertEqual(n2k(num, to="currency", currency="JPY"), out)
        with self.assertRaises(ValueError):
            n2k(190.55, to="currency")
        with self.assertRaises(NotImplementedError):
            n2k(4, to="currency", currency="EUR")

    def test_ordinal(self):
        cases = [(1, "첫 번째"), (101, "백 한 번째"), (2, "두 번째"), (5, "다섯 번째"),
                 (10, "열 번째"), (25, "스물다섯 번째"), (137, "백 서른일곱 번째")]
        for num, out in cases:
            self.assertEqual(n2k(num, to="ordinal"), out)

    def test_ordinal_num(self):
        cases = [(1, "1 번째"), (101, "101 번째"), (25, "25 번째")]
        for num, out in cases:
            self.assertEqual(n2k(num, to="ordinal_num"), out)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from . import lang_EU

GENERIC_DOLLARS = ("доллар", "доллар")
GENERIC_CENTS = ("сент", "сент")


class Num2Word_TG(lang_EU.Num2Word_EU):
    CURRENCY_FORMS = {
        # repalced by EUR
        "EUR": (("евро", "евро"), GENERIC_CENTS),
        # replaced by EUR
        "USD": (GENERIC_DOLLARS, GENERIC_CENTS),
        "RUB": (("рубл", "рубл"), ("копейк", "копейк")),
        "TJS": (("сомонӣ", "сомонӣ"), ("дирам", "дирам")),
    }

    GIGA_SUFFIX = "иллиард"
    MEGA_SUFFIX = "иллион"

    def set_high_numwords(self, high):
        cap = 3 * (len(high) + 1)

        for word, n in zip(high, range(cap, 5, -3)):
            if n == 9:
                self.cards[10 ** n] = word + self.GIGA_SUFFIX

            else:
                self.cards[10 ** n] = word + self.MEGA_SUFFIX

    def setup(self):
        super(Num2Word_TG, self).setup()

        lows = ["квинт", "квадр", "тр", "м", "м"]
        self.high_numwords = self.gen_high_numwords([], [], lows)
        self.negword = "минус "
        self.pointword = "нуқта"
        self.exclude_title = ["ва", "минус", "нуқта"]

        self.mid_numwords = [
            (1000, "ҳазор"),
            (100, "сад"),
            (90, "навад"),
            (80, "ҳаштод"),
            (70, "ҳафтод"),
            (60, "шаст"),
            (50, "панҷоҳ"),
            (40, "чил"),
            (30, "си"),
        ]
        self.low_numwords = [
            "бист",
            "нуздаҳ",
            "ҳаждаҳ",
            "ҳабдаҳ",
            "шонздаҳ",
            "понздаҳ",
            "чордаҳ",
            "сенздаҳ",
            "дувоздаҳ",
            "ёздаҳ",
            "даҳ",
            "нӯҳ",
            "ҳашт",
            "ҳафт",
            "шаш",
            "панҷ",
            "чор",
            "се",
            "ду",
            "як",
            "сифр",
        ]

    def to_cardinal(self, value):
        try:
            assert int(value) == value
        except (ValueError, TypeError, AssertionError):
            return self.to_cardinal_float(value)

        out = ""
        if value < 0:
            value = abs(value)
            out = self.negword

        if value >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (value, self.MAXVAL))

        if value == 100:
            return self.title(out + "сад")
        else:
            val = self.splitnum(value)
            words, num = self.clean(val)
            return self.title(out + words)

    def merge(self, lpair, rpair):
        ltext, lnum = lpair
        rtext, rnum = rpair
        if lnum == 1 and rnum < 100:
            return (rtext, rnum)
        elif 100 > lnum > rnum:
            if ltext == "си":
                return ("%sю %s" % (ltext, rtext), lnum + rnum)
            elif ltext == "панҷоҳ":
                return ("панҷову %s" % (rtext), lnum + rnum)
            else:
                return ("%sу %s" % (ltext, rtext), lnum + rnum)
        elif lnum >= 100 > rnum:
            return ("%sу %s" % (ltext, rtext), lnum + rnum)
        elif rnum > lnum:
            if ltext == "яксад" and rtext not in self.low_numwords:
                return ("сад %s" % (rtext), lnum * rnum)
            if rtext == "сад":
                return ("%s%s" % (ltext, rtext), lnum * rnum)
            else:
                return ("%s %s" % (ltext, rtext), lnum * rnum)
        return ("%sу %s" % (ltext, rtext), lnum + rnum)

    def to_ordinal(self, value):
        self.verify_ordinal(value)
        cardinal = self.to_cardinal(value)
        outwords = cardinal.split(" ")
        lastword = outwords[-1]
        if lastword in ["ду", "се", "си"]:
            return "%sюм" % (cardinal)
        else:
            return "%sум" % (cardinal)

    def to_ordinal_num(self, value):
        self.verify_ordinal(value)
        return "%s%s" % (value, self.to_ordinal(value)[-2:])



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA
from __future__ import unicode_literals

from unittest import TestCase

from num2words import num2words

TEST_CASES_CARDINAL = (
    (1, 'un'),
    (2, 'deux'),
    (3, 'trois'),
    (5.5, 'cinq virgule cinq'),
    (11, 'onze'),
    (12, 'douze'),
    (16, 'seize'),
    (17.42, 'dix-sept virgule quatre deux'),
    (19, 'dix-neuf'),
    (20, 'vingt'),
    (21, 'vingt et un'),
    (26, 'vingt-six'),
    (27.312, 'vingt-sept virgule trois un deux'),
    (28, 'vingt-huit'),
    (30, 'trente'),
    (31, 'trente et un'),
    (40, 'quarante'),
    (44, 'quarante-quatre'),
    (50, 'cinquante'),
    (53.486, 'cinquante-trois virgule quatre huit six'),
    (55, 'cinquante-cinq'),
    (60, 'soixante'),
    (67, 'soixante-sept'),
    (70, 'soixante-dix'),
    (79, 'soixante-dix-neuf'),
    (89, 'quatre-vingt-neuf'),
    (95, 'quatre-vingt-quinze'),
    (100, 'cent'),
    (101, 'cent un'),
    (199, 'cent quatre-vingt-dix-neuf'),
    (203, 'deux cent trois'),
    (287, 'deux cent quatre-vingt-sept'),
    (300.42, 'trois cents virgule quatre deux'),
    (356, 'trois cent cinquante-six'),
    (400, 'quatre cents'),
    (434, 'quatre cent trente-quatre'),
    (578, 'cinq cent soixante-dix-huit'),
    (689, 'six cent quatre-vingt-neuf'),
    (729, 'sept cent vingt-neuf'),
    (894, 'huit cent quatre-vingt-quatorze'),
    (999, 'neuf cent quatre-vingt-dix-neuf'),
    (1000, 'mille'),
    (1001, 'mille un'),
    (1097, 'mille quatre-vingt-dix-sept'),
    (1104, 'mille cent quatre'),
    (1243, 'mille deux cent quarante-trois'),
    (2385, 'deux mille trois cent quatre-vingt-cinq'),
    (3766, 'trois mille sept cent soixante-six'),
    (4196, 'quatre mille cent quatre-vingt-seize'),
    (4196.42, 'quatre mille cent quatre-vingt-seize virgule quatre deux'),
    (5846, 'cinq mille huit cent quarante-six'),
    (6459, 'six mille quatre cent cinquante-neuf'),
    (7232, 'sept mille deux cent trente-deux'),
    (8569, 'huit mille cinq cent soixante-neuf'),
    (9539, 'neuf mille cinq cent trente-neuf'),
    (1000000, 'un million'),
    (1000001, 'un million un'),
    (4000000, 'quatre millions'),
    (4000004, 'quatre millions quatre'),
    (4300000, 'quatre millions trois cent mille'),
    (80000000, 'quatre-vingts millions'),
    (300000000, 'trois cents millions'),
    (10000000000000, 'dix billions'),
    (10000000000010, 'dix billions dix'),
    (100000000000000, 'cent billions'),
    (1000000000000000000, 'un trillion'),
    (1000000000000000000000, 'un trilliard'),
    (10000000000000000000000000, 'dix quadrillions')
)

TEST_CASES_ORDINAL = (
    (1, 'premier'),
    (8, 'huitième'),
    (12, 'douzième'),
    (14, 'quatorzième'),
    (28, 'vingt-huitième'),
    (100, 'centième'),
    (1000, 'millième'),
    (1000000, 'un millionième'),
    (1000000000000000, 'un billiardième'),
    (1000000000000000000, 'un trillionième')  # over 1e18 is not supported
)

TEST_CASES_ORDINAL_NUM = (
    (1, '1er'),
    (8, '8me'),
    (11, '11me'),
    (12, '12me'),
    (14, '14me'),
    (21, '21me'),
    (28, '28me'),
    (100, '100me'),
    (101, '101me'),
    (1000, '1000me'),
    (1000000, '1000000me')
)

TEST_CASES_TO_CURRENCY_EUR = (
    (1.00, 'un euro et zéro centimes'),
    (2.01, 'deux euros et un centime'),
    (8.10, 'huit euros et dix centimes'),
    (12.26, 'douze euros et vingt-six centimes'),
    (21.29, 'vingt et un euros et vingt-neuf centimes'),
    (81.25, 'quatre-vingt-un euros et vingt-cinq centimes'),
    (100.00, 'cent euros et zéro centimes'),
)

TEST_CASES_TO_CURRENCY_FRF = (
    (1.00, 'un franc et zéro centimes'),
    (2.01, 'deux francs et un centime'),
    (8.10, 'huit francs et dix centimes'),
    (12.27, 'douze francs et vingt-sept centimes'),
    (21.29, 'vingt et un francs et vingt-neuf centimes'),
    (81.25, 'quatre-vingt-un francs et vingt-cinq centimes'),
    (100.00, 'cent francs et zéro centimes'),
)

TEST_CASES_TO_CURRENCY_USD = (
    (1.00, 'un dollar et zéro cents'),
    (2.01, 'deux dollars et un cent'),
    (8.10, 'huit dollars et dix cents'),
    (12.26, 'douze dollars et vingt-six cents'),
    (21.29, 'vingt et un dollars et vingt-neuf cents'),
    (81.25, 'quatre-vingt-un dollars et vingt-cinq cents'),
    (100.00, 'cent dollars et zéro cents'),
)


class Num2WordsFRTest(TestCase):
    def test_ordinal_special_joins(self):
        # ref https://github.com/savoirfairelinux/num2words/issues/18
        self.assertEqual(
            num2words(5, ordinal=True, lang='fr'), "cinquième"
        )
        self.assertEqual(
            num2words(35, ordinal=True, lang='fr'), "trente-cinquième"
        )
        self.assertEqual(
            num2words(9, ordinal=True, lang='fr'), "neuvième"
        )
        self.assertEqual(
            num2words(49, ordinal=True, lang='fr'), "quarante-neuvième"
        )

    def test_number(self):
        for test in TEST_CASES_CARDINAL:
            self.assertEqual(num2words(test[0], lang='fr'), test[1])

    def test_ordinal(self):
        for test in TEST_CASES_ORDINAL:
            self.assertEqual(
                num2words(test[0], lang='fr', ordinal=True),
                test[1]
            )

    def test_ordinal_num(self):
        for test in TEST_CASES_ORDINAL_NUM:
            self.assertEqual(
                num2words(test[0], lang='fr', to='ordinal_num'),
                test[1]
            )

    def test_currency_eur(self):
        for test in TEST_CASES_TO_CURRENCY_EUR:
            self.assertEqual(
                num2words(test[0], lang='fr', to='currency', currency='EUR'),
                test[1]
            )

    def test_currency_frf(self):
        for test in TEST_CASES_TO_CURRENCY_FRF:
            self.assertEqual(
                num2words(test[0], lang='fr', to='currency', currency='FRF'),
                test[1]
            )

    def test_currency_usd(self):
        for test in TEST_CASES_TO_CURRENCY_USD:
            self.assertEqual(
                num2words(test[0], lang='fr', to='currency', currency='USD'),
                test[1]
            )

    def test_max_numbers(self):

        with self.assertRaises(OverflowError) as context:
            num2words(10 ** 700, lang='fr')

        self.assertTrue('trop grand' in str(context.exception))



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from collections import OrderedDict

from . import lang_EU

GENERIC_CENTS = ('sentti', 'senttiä')
GENERIC_CENTAVOS = ('centavo', 'centavoa')

# grammatical cases
NOM = 10  # nominative: the dictionary form
GEN = 11  # genitive: ~of/'s
ACC = 12  # accusative: not used; either nominative or genitive
PTV = 13  # partitive: as an object
# locative cases (internal)
INE = 14  # inessive: in
ELA = 15  # elative: from/out of
ILL = 16  # illative: into
# locative cases (external)
ADE = 17  # adessive: at/on
ABL = 18  # ablative: from (after being at/on, not in)
ALL = 19  # allative: to
# essive
ESS = 20     # essive: as (in the role of)
TRANSL = 21  # translative: to (the role of; being sth)
# rare
INSTRUC = 22  # instructive: with (plural is the same as singular)
ABE = 23      # abessive: without
COM = 24      # comitative: together with (plural = singular)

NAME_TO_CASE = {
    'nominative': NOM,
    'genitive': GEN,
    'accusative': ACC,
    'partitive': PTV,
    'inessive': INE,
    'elative': ELA,
    'illative': ILL,
    'adessive': ADE,
    'ablative': ABL,
    'allative': ALL,
    'essive': ESS,
    'translative': TRANSL,
    'instructive': INSTRUC,
    'abessive': ABE,
    'comitative': COM,
}

# https://en.wikibooks.org/wiki/Finnish/Grammar-Vowel_harmony
BACK_TO_FRONT = {
    'a': 'ä',
    'o': 'ö',
    'u': 'y',
}

# https://en.wiktionary.org/wiki/Appendix:Finnish_nominal_inflection
# CASE: (SINGULAR_SUFFIX+, PLURAL_SUFFIX+)
KOTUS_TYPE = {

    # Kotus type 5/risti, no gradation
    5: {
        # grammatical
        NOM: ('i', 'it'),
        GEN: ('in', 'ien'),
        PTV: ('ia', 'eja'),
        # locative, internal
        INE: ('issa', 'eissa'),
        ELA: ('ista', 'eista'),
        ILL: ('iin', 'eihin'),
        # locative, external
        ADE: ('illa', 'eilla'),
        ABL: ('ilta', 'eilta'),
        ALL: ('ille', 'eille'),
        # essive
        ESS: ('ina', 'eina'),
        TRANSL: ('iksi', 'eiksi'),
        # rare
        INSTRUC: ('ein', 'ein'),
        ABE: ('itta', 'eitta'),
        COM: ('eine', 'eine'),  # works better
    },

    # Kotus type 7/ovi, no gradation
    7: {
        # grammatical
        NOM: ('i', 'et'),
        GEN: ('en', 'ien'),
        PTV: ('ea', 'ia'),
        # locative, internal
        INE: ('essa', 'issa'),
        ELA: ('esta', 'ista'),
        ILL: ('een', 'iin'),
        # locative, external
        ADE: ('ella', 'illa'),
        ABL: ('elta', 'ilta'),
        ALL: ('elle', 'ille'),
        # essive
        ESS: ('ena', 'ina'),
        TRANSL: ('eksi', 'iksi'),
        # rare
        INSTRUC: ('in', 'in'),
        ABE: ('etta', 'itta'),
        COM: ('ine', 'ine'),  # works better
    },

    # Kotus type 8/nalle, no gradation
    8: {
        # grammatical
        NOM: ('e', 'et'),
        GEN: ('en', ('ejen', 'ein')),
        PTV: ('ea', 'eja'),
        # locative, internal
        INE: ('essa', 'eissa'),
        ELA: ('esta', 'eista'),
        ILL: ('een', 'eihin'),
        # locative, external
        ADE: ('ella', 'eilla'),
        ABL: ('elta', 'eilta'),
        ALL: ('elle', 'eille'),
        # essive
        ESS: ('ena', 'eina'),
        TRANSL: ('eksi', 'eiksi'),
        # rare
        INSTRUC: ('ein', 'ein'),
        ABE: ('etta', 'eitta'),
        COM: ('eine', 'eine'),  # works better
    },

    # Kotus type 9/kala, t-d gradation (sata)
    109: {
        # grammatical
        NOM: ('ta', 'dat'),
        GEN: ('dan', ('tojen', 'tain')),
        PTV: ('taa', 'toja'),
        # locative, internal
        INE: ('dassa', 'doissa'),
        ELA: ('dasta', 'doista'),
        ILL: ('taan', 'toihin'),
        # locative, external
        ADE: ('dalla', 'doilla'),
        ABL: ('dalta', 'doilta'),
        ALL: ('dalle', 'doille'),
        # essive
        ESS: ('tana', 'toina'),
        TRANSL: ('daksi', 'doiksi'),
        # rare
        INSTRUC: ('doin', 'doin'),
        ABE: ('datta', 'doitta'),
        COM: ('toine', 'toine'),  # works better
    },

    # Kotus type 10/koira, no gradation
    10: {
        # grammatical
        NOM: ('a', 'at'),
        GEN: ('an', ('ien', 'ain')),
        PTV: ('aa', 'ia'),
        # locative, internal
        INE: ('assa', 'issa'),
        ELA: ('asta', 'ista'),
        ILL: ('aan', 'iin'),
        # locative, external
        ADE: ('alla', 'illa'),
        ABL: ('alta', 'ilta'),
        ALL: ('alle', 'ille'),
        # essive
        ESS: ('ana', 'ina'),
        TRANSL: ('aksi', 'iksi'),
        # rare
        INSTRUC: ('in', 'in'),
        ABE: ('atta', 'itta'),
        COM: ('ine', 'ine'),  # works better
    },

    # Kotus type 27/käsi, t-d gradation
    27: {
        # grammatical
        NOM: ('si', 'det'),
        GEN: ('den', ('sien', 'tten')),
        PTV: ('tta', 'sia'),
        # locative, internal
        INE: ('dessa', 'sissa'),
        ELA: ('desta', 'sista'),
        ILL: ('teen', 'siin'),
        # locative, external
        ADE: ('della', 'silla'),
        ABL: ('delta', 'silta'),
        ALL: ('delle', 'sille'),
        # essive
        ESS: ('tena', 'sina'),
        TRANSL: ('deksi', 'siksi'),
        # rare
        INSTRUC: ('sin', 'sin'),
        ABE: ('detta', 'sitta'),
        COM: ('sine', 'sine'),  # works better
    },

    # Kotus type 31/kaksi, t-d gradation
    31: {
        # grammatical
        NOM: ('ksi', 'hdet'),
        GEN: ('hden', 'ksien'),
        PTV: ('hta', 'ksia'),
        # locative, internal
        INE: ('hdessa', 'ksissa'),
        ELA: ('hdesta', 'ksista'),
        ILL: ('hteen', 'ksiin'),
        # locative, external
        ADE: ('hdella', 'ksilla'),
        ABL: ('hdelta', 'ksilta'),
        ALL: ('hdelle', 'ksille'),
        # essive
        ESS: ('htena', 'ksina'),
        TRANSL: ('hdeksi', 'ksiksi'),
        # rare
        INSTRUC: ('ksin', 'ksin'),
        ABE: ('hdetta', 'ksitta'),
        COM: ('ksine', 'ksine'),  # works better
    },

    # Kotus type 32/sisar, no gradation
    32: {
        # grammatical
        NOM: ('', 'et'),
        GEN: ('en', ('ien', 'ten')),
        PTV: ('ta', 'ia'),
        # locative, internal
        INE: ('essa', 'issa'),
        ELA: ('esta', 'ista'),
        ILL: ('een', 'iin'),
        # locative, external
        ADE: ('ella', 'illa'),
        ABL: ('elta', 'ilta'),
        ALL: ('elle', 'ille'),
        # essive
        ESS: ('ena', 'ina'),
        TRANSL: ('eksi', 'iksi'),
        # rare
        INSTRUC: ('in', 'in'),
        ABE: ('etta', 'itta'),
        COM: ('ine', 'ine'),  # works better
    },

    # Kotus type 38/nainen, no gradation
    38: {
        # grammatical
        NOM: ('nen', 'set'),
        GEN: ('sen', ('sten', 'sien')),
        PTV: ('sta', 'sia'),
        # locative, internal
        INE: ('sessa', 'sissa'),
        ELA: ('sesta', 'sista'),
        ILL: ('seen', 'siin'),
        # locative, external
        ADE: ('sella', 'silla'),
        ABL: ('selta', 'silta'),
        ALL: ('selle', 'sille'),
        # essive
        ESS: ('sena', 'sina'),
        TRANSL: ('seksi', 'siksi'),
        # rare
        INSTRUC: ('sin', 'sin'),
        ABE: ('setta', 'sitta'),
        COM: ('sine', 'sine'),  # works better
    },

    # Kotus type 45/kahdeksas, nt-nn gradation
    45: {
        # grammatical
        NOM: ('s', 'nnet'),
        GEN: ('nnen', 'nsien'),
        PTV: ('tta', 'nsia'),
        # locative, internal
        INE: ('nnessa', 'nsissa'),
        ELA: ('nnesta', 'nsista'),
        ILL: ('nteen', 'nsiin'),
        # locative, external
        ADE: ('nnella', 'nsilla'),
        ABL: ('nnelta', 'nsilta'),
        ALL: ('nnelle', 'nsille'),
        # essive
        ESS: ('ntena', 'nsina'),
        TRANSL: ('nneksi', 'nsiksi'),
        # rare
        INSTRUC: ('nsin', 'nsin'),
        ABE: ('nnetta', 'nsitta'),
        COM: ('nsine', 'nsine'),  # works better
    },

    # Kotus type 46/tuhat, nt-nn gradation
    46: {
        # grammatical
        NOM: ('t', 'nnet'),
        GEN: ('nnen', ('nsien', 'nten')),
        PTV: ('tta', 'nsia'),
        # locative, internal
        INE: ('nnessa', 'nsissa'),
        ELA: ('nnesta', 'nsista'),
        ILL: ('nteen', 'nsiin'),
        # locative, external
        ADE: ('nnella', 'nsilla'),
        ABL: ('nnelta', 'nsilta'),
        ALL: ('nnelle', 'nsille'),
        # essive
        ESS: ('ntena', 'nsina'),
        TRANSL: ('nneksi', 'nsiksi'),
        # rare
        INSTRUC: ('nsin', 'nsin'),
        ABE: ('nnetta', 'nsitta'),
        COM: ('nsine', 'nsine'),  # works better
    },
}

# kolme
KOTUS_TYPE[108] = {
    c: (KOTUS_TYPE[8][c][0], KOTUS_TYPE[7][c][1])
    for c in KOTUS_TYPE[8]
}
KOTUS_TYPE[108][INSTRUC] = ('en', 'in')
KOTUS_TYPE[108][ABE] = ('etta', 'itta')
KOTUS_TYPE[108][COM] = ('ine', 'ine')

# seitsemän, kahdeksan, yhdeksän
KOTUS_TYPE[110] = KOTUS_TYPE[10].copy()
KOTUS_TYPE[110][NOM] = ('an', 'at')

# kymmenen
KOTUS_TYPE[132] = KOTUS_TYPE[32].copy()
KOTUS_TYPE[132][NOM] = ('en', 'et')


def inflect(parts, options):
    if not isinstance(parts, list):
        parts = [parts]

    out = ''
    for part in parts:
        # part is plain text, concat and continue
        if not isinstance(part, tuple):
            out += part
            continue
        # predefined case (kaksikymmentä, ...)
        tmp_case = options.case
        if len(part) == 3:
            # override singular nominative only
            if options.case == NOM and not options.plural:
                tmp_case = part[2]
            part = part[:2]
        # stem and suffix
        stem, kotus_type = part
        suffix = KOTUS_TYPE[kotus_type][tmp_case][options.plural]
        # many choices, choose preferred or first
        if isinstance(suffix, tuple):
            common = set(suffix) & set(options.prefer or set())
            if len(common) == 1:
                suffix = common.pop()
            else:
                suffix = suffix[0]
        # apply vowel harmony
        if not set(BACK_TO_FRONT) & set(stem):
            for back, front in BACK_TO_FRONT.items():
                suffix = suffix.replace(back, front)
        # concat
        out += stem + suffix

    return out


class Options(object):
    def __init__(self, ordinal, case, plural, prefer):
        self.ordinal = ordinal
        self.case = case
        self.plural = plural
        self.prefer = prefer

    def variation(self, ordinal=None, case=None, plural=None, prefer=None):
        return Options(
            ordinal if ordinal is not None else self.ordinal,
            case if case is not None else self.case,
            plural if plural is not None else self.plural,
            prefer if prefer is not None else self.prefer,
        )


class Num2Word_FI(lang_EU.Num2Word_EU):
    CURRENCY_FORMS = {
        'BRL': (('real', 'realia'), GENERIC_CENTAVOS),
        'CHF': (('frangi', 'frangia'), ('rappen', 'rappenia')),
        'CNY': (('juan', 'juania'), ('fen', 'feniä')),
        'EUR': (('euro', 'euroa'), GENERIC_CENTS),
        'FIM': (('markka', 'markkaa'), ('penni', 'penniä')),  # historical
        'INR': (('rupia', 'rupiaa'), ('paisa', 'paisaa')),
        'JPY': (('jeni', 'jeniä'), ('sen', 'seniä')),  # rare subunit
        'KRW': (('won', 'wonia'), ('jeon', 'jeonia')),  # rare subunit
        'KPW': (('won', 'wonia'), ('chon', 'chonia')),  # rare subunit
        'MXN': (('peso', 'pesoa'), GENERIC_CENTAVOS),
        'RUB': (('rupla', 'ruplaa'), ('kopeekka', 'kopeekkaa')),
        'TRY': (('liira', 'liiraa'), ('kuruş', 'kuruşia')),
        'ZAR': (('randi', 'randia'), GENERIC_CENTS),
    }

    # crowns
    for curr_code in 'DKK', 'ISK', 'NOK', 'SEK':
        CURRENCY_FORMS[curr_code] = (('kruunu', 'kruunua'), ('äyri', 'äyriä'))

    # dollars
    for curr_code in 'AUD', 'CAD', 'HKD', 'NZD', 'SGD', 'USD':
        CURRENCY_FORMS[curr_code] = (
            ('dollari', 'dollaria'), GENERIC_CENTS)

    # pounds
    for curr_code in ('GBP',):
        CURRENCY_FORMS[curr_code] = (('punta', 'puntaa'), ('penny', 'pennyä'))

    CURRENCY_ADJECTIVES = {
        'AUD': 'Australian',
        'BRL': 'Brasilian',
        'CAD': 'Kanadan',
        'CHF': 'Sveitsin',
        'DKK': 'Tanskan',
        'FIM': 'Suomen',  # historical
        'GBP': 'Englannin',
        'HKD': 'Hongkongin',
        'INR': 'Intian',
        'ISK': 'Islannin',
        'KRW': 'Etelä-Korean',
        'KPW': 'Pohjois-Korean',
        'MXN': 'Meksikon',
        'NOK': 'Norjan',
        'NZD': 'Uuden-Seelannin',
        'RUB': 'Venäjän',
        'SEK': 'Ruotsin',
        'SGD': 'Singaporen',
        'TRY': 'Turkin',
        'USD': 'Yhdysvaltain',
        'ZAR': 'Etelä-Afrikan',
    }

    def __init__(self):
        self.ords = OrderedDict()
        super(Num2Word_FI, self).__init__()

    def set_numwords(self):
        self.set_high_numwords(self.high_numwords)
        self.set_mid_numwords(self.mid_numwords, self.mid_ords)
        self.set_low_numwords(self.low_numwords, self.low_ords)

    def set_high_numwords(self, high):
        # references:
        # https://fi.wikipedia.org/wiki/Suurten_lukujen_nimet
        # https://en.wikipedia.org/wiki/Names_of_large_numbers#Standard_dictionary_numbers

        # translate to Finnish
        replacements = [
            ("qu", "kv"),
            ("x", "ks"),
            ("c", "k"),
            ("kent", "sent"),  # applied after c -> k to cent
        ]
        translated = []
        for i, numword in enumerate(high):
            # notes:
            # - 1e6**9 can be either noviljoona or noniljoona
            # - 1e6**38 and above are untested

            # 1e6**6 is sekstiljoona but 1e6**16 is sedekiljoona
            if numword.startswith("sex") and numword != "sext":
                numword = numword.replace("sex", "se")
            # 1e6**7 is septiljoona but 1e6**17 is septendekiljoona
            elif numword.startswith("sept") and numword != "sept":
                numword = "septen" + numword[len("sept"):]
            # 1e6**8 is oktiljoona but 1e6**18 is duodevigintiljoona
            # (2 from 20)
            elif numword.startswith("octo"):
                numword = high[i + -10]
                numword = "duode" + numword[len("octo"):]
            # 1e6**9 is noniljoona but 1e6**19 is undevigintiljoona (1 from 20)
            elif numword.startswith("nove"):
                numword = high[i + -10]
                numword = "unde" + numword[len("nove") + 1:]

            # apply general replacements to all numwords
            for repl in replacements:
                numword = numword.replace(repl[0], repl[1])
            translated.append(numword)

        max = 6 * len(translated)
        for word, n in zip(translated, range(max, 0, -6)):
            if n == 6:
                # irregularity considering short scale and long scale
                self.cards[10 ** 9] = ("miljard", 5)
                self.ords[10 ** 9] = ("miljardi", 45)
            self.cards[10 ** n] = (word + "iljoon", 10)
            self.ords[10 ** n] = (word + "iljoona", 45)

    def set_mid_numwords(self, cards, ords):
        for key, val in cards:
            self.cards[key] = val
        for key, val in ords:
            self.ords[key] = val

    def set_low_numwords(self, cards, ords):
        for key, val in cards:
            self.cards[key] = val
        for key, val in ords:
            self.ords[key] = val

    def setup(self):
        super(Num2Word_FI, self).setup()

        self.negword = "miinus "
        self.pointword = "pilkku"
        self.exclude_title = ["pilkku", "miinus"]

        self.mid_numwords = [
            (1000, ("tuha", 46)),
            (100, ("sa", 109)),
            (90, [("yhdeks", 110), ("kymmen", 132, PTV)]),
            (80, [("kahdeks", 110), ("kymmen", 132, PTV)]),
            (70, [("seitsem", 110), ("kymmen", 132, PTV)]),
            (60, [("kuu", 27), ("kymmen", 132, PTV)]),
            (50, [("vii", 27), ("kymmen", 132, PTV)]),
            (40, [("nelj", 10), ("kymmen", 132, PTV)]),
            (30, [("kolm", 108), ("kymmen", 132, PTV)]),
        ]

        self.mid_ords = [
            (1000, ("tuhanne", 45)),
            (100, ("sada", 45)),
            (90, [("yhdeksä", 45), ("kymmene", 45)]),
            (80, [("kahdeksa", 45), ("kymmene", 45)]),
            (70, [("seitsemä", 45), ("kymmene", 45)]),
            (60, [("kuude", 45), ("kymmene", 45)]),
            (50, [("viide", 45), ("kymmene", 45)]),
            (40, [("neljä", 45), ("kymmene", 45)]),
            (30, [("kolma", 45), ("kymmene", 45)]),
        ]

        self.low_numwords = [
            (20, [("ka", 31), ("kymmen", 132, PTV)]),
            (19, [("yhdeks", 110), "toista"]),
            (18, [("kahdeks", 110), "toista"]),
            (17, [("seitsem", 110), "toista"]),
            (16, [("kuu", 27), "toista"]),
            (15, [("vii", 27), "toista"]),
            (14, [("nelj", 10), "toista"]),
            (13, [("kolm", 108), "toista"]),
            (12, [("ka", 31), "toista"]),
            (11, [("y", 31), "toista"]),
            (10, ("kymmen", 132)),
            (9, ("yhdeks", 110)),
            (8, ("kahdeks", 110)),
            (7, ("seitsem", 110)),
            (6, ("kuu", 27)),
            (5, ("vii", 27)),
            (4, ("nelj", 10)),
            (3, ("kolm", 108)),
            (2, ("ka", 31)),
            (1, ("y", 31)),
            (0, ("noll", 10)),
        ]

        self.low_ords = [
            (20, [("kahde", 45), ("kymmene", 45)]),
            (19, [("yhdeksä", 45), "toista"]),
            (18, [("kahdeksa", 45), "toista"]),
            (17, [("seitsemä", 45), "toista"]),
            (16, [("kuude", 45), "toista"]),
            (15, [("viide", 45), "toista"]),
            (14, [("neljä", 45), "toista"]),
            (13, [("kolma", 45), "toista"]),
            (12, [("kahde", 45), "toista"]),
            (11, [("yhde", 45), "toista"]),
            (10, ("kymmene", 45)),
            (9, ("yhdeksä", 45)),
            (8, ("kahdeksa", 45)),
            (7, ("seitsemä", 45)),
            (6, ("kuude", 45)),
            (5, ("viide", 45)),
            (4, ("neljä", 45)),
            (3, ("kolma", 45)),
            (2, ("toi", 38)),
            (1, ("ensimmäi", 38)),
            (0, ("nolla", 45)),
        ]

    def merge(self, lpair, rpair, options):
        ltext, lnum = lpair
        rtext, rnum = rpair

        # http://www.kielitoimistonohjepankki.fi/ohje/49
        fmt = "%s%s"
        # ignore lpair if lnum is 1
        if lnum == 1:
            rtext = inflect(rtext, options)
            return (rtext, rnum)
        # rnum is added to lnum
        elif lnum > rnum:
            ltext = inflect(ltext, options)
            rtext = inflect(rtext, options)
            # separate groups with space
            if lnum >= 1000:
                fmt = "%s %s"
            return (fmt % (ltext, rtext), lnum + rnum)
        # rnum is multiplied by lnum
        elif lnum < rnum:
            if options.ordinal:
                # kahdessadas, not toinensadas
                if lnum == 2:
                    ltext = ("kahde", 45)
                rtext = inflect(rtext, options)
            else:
                # kaksituhatta but kahdettuhannet
                rcase = options.case
                if options.case == NOM and not options.plural:
                    rcase = PTV
                rtext = inflect(rtext, options.variation(case=rcase))
            ltext = inflect(ltext, options)
            return (fmt % (ltext, rtext), lnum * rnum)

    def to_cardinal(self, value, case='nominative', plural=False, prefer=None):
        case = NAME_TO_CASE[case]
        options = Options(False, case, plural, prefer)
        try:
            assert int(value) == value
        except (ValueError, TypeError, AssertionError):
            if case != NOM:
                raise NotImplementedError(
                    "Cases other than nominative are not implemented for "
                    "cardinal floating point numbers.")
            return self.to_cardinal_float(value)

        out = ""
        if value < 0:
            value = abs(value)
            out = self.negword

        if value >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (value, self.MAXVAL))

        val = self.splitnum(value, options)
        words, num = self.clean(val, options)
        return self.title(out + words)

    def to_ordinal(self, value, case='nominative', plural=False, prefer=None):
        case = NAME_TO_CASE[case]
        options = Options(True, case, plural, prefer)

        self.verify_ordinal(value)
        if value >= self.MAXVAL:
            raise OverflowError(self.errmsg_toobig % (value, self.MAXVAL))

        val = self.splitnum(value, options)
        words, num = self.clean(val, options)
        return self.title(words)

    def to_ordinal_num(self, value, case='nominative', plural=False):
        case = NAME_TO_CASE[case]
        raise NotImplementedError

    def to_year(self, val, suffix=None, longval=True):
        suffix = suffix or ""
        if val < 0:
            val = abs(val)
            suffix = suffix or " ennen ajanlaskun alkua"
        return self.to_cardinal(val).replace(" ", "") + suffix

    def to_currency(self, val, currency="EUR", cents=True, separator=" ja",
                    adjective=False):
        return super(Num2Word_FI, self).to_currency(
            val, currency=currency, cents=cents, separator=separator,
            adjective=adjective)

    def splitnum(self, value, options):
        elems = self.ords if options.ordinal else self.cards
        for elem in elems:
            if elem > value:
                continue

            out = []
            if value == 0:
                div, mod = 1, 0
            else:
                div, mod = divmod(value, elem)

            if div == 1:
                out.append((elems[1], 1))
            else:
                if div == value:  # The system tallies, eg Roman Numerals
                    return [(div * elems[elem], div*elem)]
                out.append(self.splitnum(div, options))

            out.append((elems[elem], elem))

            if mod:
                out.append(self.splitnum(mod, options))

            return out

    def clean(self, val, options):
        out = val
        while len(val) != 1:
            out = []
            left, right = val[:2]
            if isinstance(left, tuple) and isinstance(right, tuple):
                out.append(self.merge(left, right, options))
                if val[2:]:
                    out.append(val[2:])
            else:
                for elem in val:
                    if isinstance(elem, list):
                        if len(elem) == 1:
                            out.append(elem[0])
                        else:
                            out.append(self.clean(elem, options))
                    else:
                        out.append(elem)
            val = out
        return out[0]



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

import re
from io import open

from setuptools import find_packages, setup

PACKAGE_NAME = "num2words"

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU Library or Lesser General Public License '
    '(LGPL)',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Topic :: Software Development :: Internationalization',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Localization',
    'Topic :: Text Processing :: Linguistic',
]

LONG_DESC = open('README.rst', 'rt', encoding="utf-8").read() + '\n\n' + \
            open('CHANGES.rst', 'rt', encoding="utf-8").read()


def find_version(fname):
    """Parse file & return version number matching 0.0.1 regex
    Returns str or raises RuntimeError
    """
    version = ''
    with open(fname, 'r', encoding="utf-8") as fp:
        reg = re.compile(r'__version__ = [\'"]([^\'"]*)[\'"]')
        for line in fp:
            m = reg.match(line)
            if m:
                version = m.group(1)
                break
    if not version:
        raise RuntimeError('Cannot find version information')
    return version


setup(
    name=PACKAGE_NAME,
    version=find_version("bin/num2words"),
    description='Modules to convert numbers to words. Easily extensible.',
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    license='LGPL',
    author='Taro Ogawa <tso at users sourceforge net>',
    author_email='tos@users.sourceforge.net',
    maintainer='Savoir-faire Linux inc.',
    maintainer_email='support@savoirfairelinux.com',
    keywords=' number word numbers words convert conversion i18n '
             'localisation localization internationalisation '
             'internationalization',
    url='https://github.com/savoirfairelinux/num2words',
    packages=find_packages(exclude=['tests']),
    test_suite='tests',
    classifiers=CLASSIFIERS,
    scripts=['bin/num2words'],
    install_requires=["docopt>=0.6.2"],
    tests_require=['delegator.py'],
)



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import division, print_function, unicode_literals

from unittest import TestCase

from num2words import num2words


def n2j(*args, **kwargs):
    return num2words(*args, lang='ja', **kwargs)


class Num2WordsJATest(TestCase):
    def test_low(self):
        self.assertEqual(n2j(0), "零")
        self.assertEqual(n2j(0, prefer=["〇"]), "〇")
        self.assertEqual(n2j(0, reading=True), "ゼロ")
        self.assertEqual(n2j(0, reading=True, prefer=["れい"]), "れい")
        self.assertEqual(n2j(1), "一")
        self.assertEqual(n2j(1, reading=True), "いち")
        self.assertEqual(n2j(2), "二")
        self.assertEqual(n2j(2, reading=True), "に")
        self.assertEqual(n2j(3), "三")
        self.assertEqual(n2j(3, reading=True), "さん")
        self.assertEqual(n2j(4), "四")
        self.assertEqual(n2j(4, reading=True), "よん")
        self.assertEqual(n2j(4, reading=True, prefer=["し"]), "し")
        self.assertEqual(n2j(5), "五")
        self.assertEqual(n2j(5, reading=True), "ご")
        self.assertEqual(n2j(6), "六")
        self.assertEqual(n2j(6, reading=True), "ろく")
        self.assertEqual(n2j(7), "七")
        self.assertEqual(n2j(7, reading=True), "なな")
        self.assertEqual(n2j(7, reading=True, prefer=["しち"]), "しち")
        self.assertEqual(n2j(8), "八")
        self.assertEqual(n2j(8, reading=True), "はち")
        self.assertEqual(n2j(9), "九")
        self.assertEqual(n2j(9, reading=True), "きゅう")
        self.assertEqual(n2j(10), "十")
        self.assertEqual(n2j(10, reading=True), "じゅう")
        self.assertEqual(n2j(11), "十一")
        self.assertEqual(n2j(11, reading=True), "じゅういち")
        self.assertEqual(n2j(12), "十二")
        self.assertEqual(n2j(12, reading=True), "じゅうに")
        self.assertEqual(n2j(13), "十三")
        self.assertEqual(n2j(13, reading=True), "じゅうさん")
        self.assertEqual(n2j(14), "十四")
        self.assertEqual(n2j(14, reading=True), "じゅうよん")
        self.assertEqual(n2j(14, reading=True, prefer=["し"]), "じゅうし")
        self.assertEqual(n2j(15), "十五")
        self.assertEqual(n2j(15, reading=True), "じゅうご")
        self.assertEqual(n2j(16), "十六")
        self.assertEqual(n2j(16, reading=True), "じゅうろく")
        self.assertEqual(n2j(17), "十七")
        self.assertEqual(n2j(17, reading=True), "じゅうなな")
        self.assertEqual(n2j(17, reading=True, prefer=["しち"]), "じゅうしち")
        self.assertEqual(n2j(18), "十八")
        self.assertEqual(n2j(18, reading=True), "じゅうはち")
        self.assertEqual(n2j(19), "十九")
        self.assertEqual(n2j(19, reading=True), "じゅうきゅう")
        self.assertEqual(n2j(20), "二十")
        self.assertEqual(n2j(20, reading=True), "にじゅう")

    def test_mid(self):
        self.assertEqual(n2j(100), "百")
        self.assertEqual(n2j(100, reading=True), "ひゃく")
        self.assertEqual(n2j(123), "百二十三")
        self.assertEqual(n2j(123, reading=True), "ひゃくにじゅうさん")
        self.assertEqual(n2j(300), "三百")
        self.assertEqual(n2j(300, reading=True), "さんびゃく")
        self.assertEqual(n2j(400), "四百")
        self.assertEqual(n2j(400, reading=True), "よんひゃく")
        # 400 --> しひゃく sounds weird, but can be generated with prefer
        self.assertEqual(n2j(600), "六百")
        self.assertEqual(n2j(600, reading=True), "ろっぴゃく")
        self.assertEqual(n2j(700, reading=True, prefer=["しち"]), "しちひゃく")
        self.assertEqual(n2j(800, reading=True), "はっぴゃく")
        self.assertEqual(n2j(1000), "千")
        self.assertEqual(n2j(1000, reading=True), "せん")
        self.assertEqual(n2j(3000, reading=True), "さんぜん")
        self.assertEqual(n2j(8000, reading=True), "はっせん")

    def test_high(self):
        self.assertEqual(n2j(10000), "一万")
        self.assertEqual(n2j(10000, reading=True), "いちまん")
        self.assertEqual(n2j(12345), "一万二千三百四十五")
        self.assertEqual(n2j(12345, reading=True),
                         "いちまん"
                         "にせん"
                         "さんびゃく"
                         "よんじゅうご")
        self.assertEqual(n2j(10**8), "一億")
        self.assertEqual(n2j(10**8, reading=True), "いちおく")
        self.assertEqual(n2j(123456789), "一億二千三百四十五万六千七百八十九")
        self.assertEqual(n2j(123456789, reading=True),
                         "いちおく"
                         "にせんさんびゃくよんじゅうごまん"
                         "ろくせんななひゃく"
                         "はちじゅうきゅう")
        self.assertEqual(n2j(10**12), "一兆")
        self.assertEqual(n2j(10**12, reading=True), "いっちょう")
        self.assertEqual(n2j(1234567890123),
                         "一兆二千三百四十五億六千七百八十九万百二十三")
        self.assertEqual(n2j(1234567890123, reading=True),
                         "いっちょう"
                         "にせんさんびゃくよんじゅうごおく"
                         "ろくせんななひゃくはちじゅうきゅうまん"
                         "ひゃくにじゅうさん")
        # TODO: tests for 10**16 and above

    def test_cardinal_float(self):
        self.assertEqual(n2j(0.0123456789, prefer=["〇"]),
                         "〇点〇一二三四五六七八九")
        self.assertEqual(n2j(0.0123456789, reading=True),
                         "れいてん"
                         "れいいち"
                         "にさん"
                         "よんご"
                         "ろくなな"
                         "はちきゅう")
        self.assertEqual(n2j(10**8 + 0.01), "一億点零一")
        self.assertEqual(n2j(10**8 + 0.01, reading=True),
                         "いちおくてんれいいち")

    def test_ordinal(self):
        self.assertEqual(n2j(0, to="ordinal"), "零番目")
        self.assertEqual(n2j(0, to="ordinal", reading=True, prefer=["れい"]),
                         "れいばんめ")
        self.assertEqual(n2j(2, to="ordinal", counter="人"), "二人目")
        self.assertEqual(n2j(3, to="ordinal", counter="つ"), "三つ目")
        with self.assertRaises(NotImplementedError):
            n2j(4, to="ordinal", reading=True, counter="人")

    def test_ordinal_num(self):
        self.assertEqual(n2j(0, to="ordinal_num"), "0番目")
        self.assertEqual(n2j(0, to="ordinal_num", reading=True), "0ばんめ")
        self.assertEqual(n2j(2, to="ordinal_num", counter="人"), "2人目")
        self.assertEqual(n2j(3, to="ordinal_num", counter="つ"), "3つ目")

    def test_currency(self):
        self.assertEqual(n2j(123456789, to="currency"),
                         "一億二千三百四十五万六千七百八十九円")
        self.assertEqual(n2j(123456789, to="currency", reading=True),
                         "いちおく"
                         "にせんさんびゃくよんじゅうごまん"
                         "ろくせんななひゃく"
                         "はちじゅうきゅうえん")

    def test_year(self):
        self.assertEqual(n2j(2021, to="year"), "令和三年")
        self.assertEqual(n2j(2021, to="year", reading=True),
                         "れいわさんねん")
        self.assertEqual(n2j(2021, to="year", reading="arabic"),
                         "令和3年")
        self.assertEqual(n2j(2019, to="year"), "令和元年")
        self.assertEqual(n2j(2019, to="year", reading=True),
                         "れいわがんねん")
        self.assertEqual(n2j(2019, to="year", reading="arabic"),
                         "令和1年")
        self.assertEqual(n2j(2018, to="year"), "平成三十年")
        self.assertEqual(n2j(2018, to="year", reading=True),
                         "へいせいさんじゅうねん")
        self.assertEqual(n2j(2018, to="year", reading="arabic"),
                         "平成30年")
        self.assertEqual(n2j(2017, to="year"), "平成二十九年")
        self.assertEqual(n2j(2017, to="year", reading=True),
                         "へいせいにじゅうくねん")
        self.assertEqual(n2j(2017, to="year", reading="arabic"),
                         "平成29年")
        self.assertEqual(n2j(2009, to="year", era=False), "二千九年")
        self.assertEqual(n2j(2009, to="year", reading=True, era=False),
                         "にせんくねん")
        self.assertEqual(n2j(2000, to="year", era=False), "二千年")
        self.assertEqual(n2j(2000, to="year", era=False, reading=True),
                         "にせんねん")
        self.assertEqual(n2j(645, to="year"), "大化元年")
        self.assertEqual(n2j(645, to="year", reading=True), "たいかがんねん")
        self.assertEqual(n2j(-99, to="year", era=False), "紀元前九十九年")
        self.assertEqual(n2j(-99, to="year", era=False, reading=True),
                         "きげんぜんきゅうじゅうくねん")
        self.assertEqual(n2j(1375, to="year"), "天授元年")
        self.assertEqual(n2j(1375, to="year", prefer=["えいわ"]), "永和元年")



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from __future__ import unicode_literals

from num2words.base import Num2Word_Base
from num2words.currency import parse_currency_parts
from num2words.utils import splitbyx


class Num2Word_TH(Num2Word_Base):

    def setup(self):
        self.negword = 'ติดลบ'
        self.pointword = 'จุด'

        self.CURRENCY_FORMS = {
            'THB': (('บาท', 'บาท'), ('สตางค์', 'สตางค์')),
            'USD': (('ดอลลาร์', 'ดอลลาร์'), ('เซนต์', 'เซนต์')),
            'EUR': (('ยูโร', 'ยูโร'), ('เซนต์', 'เซนต์')),
        }

        self.high_numwords = []

        self.mid_numwords = ['', 'สิบ', 'ร้อย', 'พัน', 'หมื่น', 'แสน', 'ล้าน']

        self.low_numwords = [
            'ศูนย์', 'หนึ่ง', 'สอง', 'สาม', 'สี่',
            'ห้า', 'หก', 'เจ็ด', 'แปด', 'เก้า'
        ]

    def set_high_numwords(self, high_numwords):
        pass

    def set_mid_numwords(self, mid_numwords):
        pass

    def splitnum(self, six_num):
        length = len(six_num) > 1
        word_num = ''

        for index, num in enumerate(map(int, six_num)):
            if num:
                if index:
                    word_num = self.mid_numwords[index] + word_num

                if length and num == 1 and index == 0:
                    word_num += 'เอ็ด'
                elif index == 1 and num == 2:
                    word_num = 'ยี่' + word_num
                elif index != 1 or num != 1:
                    word_num = self.low_numwords[num] + word_num

            elif num == 0 and index == 0 and length == 0:
                word_num = self.low_numwords[0]

        return word_num

    def split_six(self, num_txt):
        result = splitbyx(num_txt, 6, format_int=False)
        result = list(result)[::-1]
        number_list = []
        for i in result:
            number_list.append(i[::-1])
        return number_list

    def add_text_million(self, word_num):
        result = ''

        for index, t in enumerate(reversed(word_num)):
            if index == 0:
                result = t
            else:
                result = result + 'ล้าน' + t

        return result

    def round_2_decimal(self, number):
        integer, cents, negative = parse_currency_parts(
            number, is_int_with_cents=False
        )
        integer = '{}'.format(integer)
        cents = '{}'.format(cents)

        if len(cents) < 2:
            add_zero = 2 - len(cents)
            cents = ('0' * add_zero) + cents

        text_num = integer + '.' + cents

        return text_num, negative

    def left_num_to_text(self, number):

        left_num_list = self.split_six(number)

        left_text_list = []
        for i in left_num_list:
            left_text_list.append(self.splitnum(i))

        left_text = self.add_text_million(left_text_list)
        return left_text

    def to_cardinal(self, number):
        negative = number < 0

        pre, post = self.float2tuple(number)
        precision = self.precision
        pre = '{}'.format(pre)
        post = '{}'.format(post)

        if negative:
            pre = pre.lstrip('-')

        if len(post) < precision:
            add_zero = precision - len(post)
            post = ('0' * add_zero) + post

        result = self.left_num_to_text(pre)

        right_text = ''
        if not post == '0':
            for i in map(int, post):
                right_text = right_text + self.low_numwords[i]
            result = result + 'จุด' + right_text

        if negative:
            result = 'ติดลบ' + result

        return result

    def to_ordinal(self, number):
        return self.to_cardinal(number)

    def to_currency(self, number, currency='THB'):

        number, negative = self.round_2_decimal(number)

        split_num = number.split('.')

        left_num = split_num[0]
        left_text = self.left_num_to_text(left_num)

        right_num = split_num[1]
        right_text = self.splitnum(right_num[::-1].rstrip('0'))

        try:
            cr1, cr2 = self.CURRENCY_FORMS[currency]

        except KeyError:
            raise NotImplementedError(
                'Currency code "%s" not implemented for "%s"' %
                (currency, self.__class__.__name__))

        if right_num == '00':
            if currency == 'THB':
                result = left_text + cr1[0] + 'ถ้วน'
            else:
                result = left_text + cr1[0]
        else:
            if left_num == '0':
                result = right_text + cr2[0]
            else:
                result = left_text + cr1[0] + right_text + cr2[0]

        if negative:
            result = self.negword + result

        return result



# -*- coding: utf-8 -*-
# Copyright (c) 2003, Taro Ogawa.  All Rights Reserved.
# Copyright (c) 2013, Savoir-faire Linux inc.  All Rights Reserved.

# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301 USA

from unittest import TestCase

from num2words import num2words


class Num2WordsENINTest(TestCase):
    def test_cardinal(self):
        self.assertEqual(num2words(1e5, lang="en_IN"), "one lakh")
        self.assertEqual(num2words(1e6, lang="en_IN"), "ten lakh")
        self.assertEqual(num2words(1e7, lang="en_IN"), "one crore")
